from .. import Provider as InternetProvider


class Provider(InternetProvider):

    free_email_domains = (
        'gmail.com',
        'yahoo.com',
        'hotmail.com',
        'yahoo.com.au',
        'hotmail.com.au',
    )

    tlds = ('com', 'com.au', 'org', 'org.au', 'net',
            'net.au', 'biz', 'info', 'edu', 'edu.au')



try:
    from ..cms import renders
except ValueError:
    from cms import renders

from . import settings


class AssetRenderer(renders.ChoicesRender):
    def get_object_list(self, adm_list):
        l = []
        for row in adm_list:
            o = row.instance

            if hasattr(o.file, "admin_url"):
                thumbnail = o.file.admin_url()

            data = {"id": o.pk, "user_filename": o.user_filename, "url": o.file.url}
            if thumbnail:
                data["thumbnail"] = thumbnail

            l.append(data)

        return l

    def get_fields(self, fields):
        data = {
            "user_filename": {
                "sortable": False,
                "order_type": "",
                "name": "User Filename",
            },
            "thumbnail": {"sortable": False, "order_type": "", "name": "Thumbnail"},
            "url": {"sortable": False, "order_type": "", "name": "Url"},
        }
        return data, []



class TomlError(RuntimeError):
    def __init__(self, message, line, col, filename):
        RuntimeError.__init__(self, message, line, col, filename)
        self.message = message
        self.line = line
        self.col = col
        self.filename = filename

    def __str__(self):
        return '{}({}, {}): {}'.format(self.filename, self.line, self.col, self.message)

    def __repr__(self):
        return 'TomlError({!r}, {!r}, {!r}, {!r})'.format(self.message, self.line, self.col, self.filename)



from django.apps import apps as djapps
from django.core.exceptions import ImproperlyConfigured

from . import settings


def get_page_model():
    try:
        app_label, model_name = settings.PAGE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "PAGE_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_hero_model():
    try:
        app_label, model_name = settings.HERO_MODULE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "HERO_MODULE_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_header_model():
    try:
        app_label, model_name = settings.HEADER_MODULE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "HEADER_MODULE_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_two_column_model():
    try:
        app_label, model_name = settings.TWO_COLUMN_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "TWO_COLUMN_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_icon_list_model():
    try:
        app_label, model_name = settings.ICON_LIST_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "ICON_LIST_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_icon_item_model():
    try:
        app_label, model_name = settings.ICON_LIST_ITEM_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "ICON_LIST_ITEM_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_faq_model():
    try:
        app_label, model_name = settings.FAQ_MODULE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "FAQ_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_faq_item_model():
    try:
        app_label, model_name = settings.FAQ_ITEM_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "FAQ_ITEM_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_location_model():
    try:
        app_label, model_name = settings.LOCATION_MODULE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "LOCATION_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_location_item_model():
    try:
        app_label, model_name = settings.LOCATION_ITEM_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "LOCATION_ITEM_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_image_gallery_model():
    try:
        app_label, model_name = settings.GALLERY_MODULE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "IMAGE_GALLERY_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_gallery_item_model():
    try:
        app_label, model_name = settings.GALLERY_ITEM_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "GALLERY_ITEM_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_carousel_model():
    try:
        app_label, model_name = settings.CAROUSEL_MODULE_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "CAROUSEL_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_carousel_item_model():
    try:
        app_label, model_name = settings.CAROUSEL_ITEM_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "CAROUSEL_ITEM_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)



import copy
import sys
from collections import OrderedDict

from django import dispatch
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.fields import Field, FieldDoesNotExist, related
from django.utils import formats, timezone

try:
    from ..scheduling.models import Schedulable
except ValueError:
    from scheduling.models import Schedulable


try:
    from django.db.transaction import atomic as xact
except ImportError:
    from .transactions import xact

from . import manager


class Cloneable(models.Model):
    """
    Abstract model that knows how to clone itself.

    All version models automatically implement this model.

    If you want to clone reverse relations
    you should add the attribute names of the
    reverse relations to the _clone_related class
    attribute. Any reverse relations must also
    implement Cloneable.

    This model adds the following field:

    **last_save:** Datetime that gets updated every time the object gets saved.
    """

    last_save = models.DateTimeField(editable=False)

    def _gather_m2ms(self):
        old_m2ms = {}
        for field in self._meta.local_many_to_many:
            if (
                field.remote_field.through
                and field.remote_field.through._meta.auto_created
                and field.name not in self.clone_related
            ):
                man = getattr(self, field.name)
                l = list(man.all())
                old_m2ms[field.name] = l
        return old_m2ms

    def _get_field_map(self):
        try:
            cache = self._meta._name_map
        except AttributeError:
            try:
                cache = self._meta.init_name_map()
            except AttributeError:
                cache = {
                    field.name: self._meta.get_field(field.name)
                    for field in self._meta.get_fields()
                }
        return cache

    def _gather_reverse(self, reverse):
        cache = self._get_field_map()
        rel = cache[reverse]
        m2m = rel.is_relation and rel.many_to_many

        if m2m:
            ctype = "m2m"
            name = reverse
            manager = getattr(self, name)
        else:
            ctype = "reverse"
            name = rel.field.name
            manager = getattr(self, rel.get_accessor_name())

        l = list(manager.all())
        return ctype, name, l

    def _gather_reverses(self):
        """
        Get all the related objects that point to this
        object that we need to clone. Uses self.clone_related
        to find those objects.
        """
        old_reverses = {"m2m": {}, "reverse": {}}
        for reverse in self.clone_related:
            ctype, name, l = self._gather_reverse(reverse)
            old_reverses[ctype][reverse] = (name, l)

        return old_reverses

    def _set_m2ms(self, old_m2ms):
        """
        Creates the same m2m relationships that the old
        object had.
        """
        for k, v in list(old_m2ms.items()):
            if v:
                getattr(self, k).set(v)

    def _clone_reverses(self, old_reverses):
        """
        Clones all the objects that were previously gathered.
        """
        for ctype, reverses in list(old_reverses.items()):
            for parts in list(reverses.values()):
                sub_objs = parts[1]
                field_name = parts[0]

                attrs = {}
                for sub_obj in sub_objs:
                    if ctype != "m2m" and not attrs:
                        field = sub_obj._meta.get_field(field_name)
                        attrs = {
                            field.column: getattr(self, field.remote_field.field_name)
                        }
                    sub_obj._clone(**attrs)

                if ctype == "m2m":
                    getattr(self, field_name).set(sub_objs)

    def prep_for_clone(self):
        """
        Hook so implementing classes can customize
        what gets cleared before a clone. By default
        the pk is set to None.
        """
        self.pk = None

    def _clone(self, *args, **kwargs):
        """
        Makes a copy of an model instance.

        for every key in **kwargs value will
        be set on the new instance.
        """

        with xact():
            # Gather objs we'll need save after
            old_m2ms = self._gather_m2ms()
            old_reverses = self._gather_reverses()

            for k, v in list(kwargs.items()):
                setattr(self, k, v)

            # Do the clone
            self.prep_for_clone()
            self.validate_unique()
            # Prevent last save from changing
            try:
                self.save(last_save=self.last_save)
            except ValueError:
                # Prior to Django 1.8 it was allowed to save the _id field of an object
                # without the object itself being saved yet, but now that raises a
                # ValueError. So we are reacting to the error by looking up all FK
                # objects in the database and setting them prior to save.
                try:
                    # Needed to avoid circular import.
                    from . import fields
                except ImportError:
                    import fields
                fks = [
                    f.name
                    for f in self._meta.get_fields()
                    if isinstance(f, fields.FKToVersion)
                ]
                for fk in fks:
                    setattr(
                        self,
                        fk,
                        type(getattr(self, fk)).objects.get(vid=kwargs.get(fk + "_id")),
                    )
                self.save(last_save=self.last_save)
            # save m2ms
            self._set_m2ms(old_m2ms)
            # Prevent last save from changing
            self.save(last_save=self.last_save)
            # save reverses
            self._clone_reverses(old_reverses)

    def _delete_reverses(self):
        """
        Delete all objects that would have been cloned
        on a clone command. This is done separately because
        there may be m2m and other relationships that
        would have not been deleted otherwise.
        """

        for reverse in self.clone_related:
            self._delete_reverse(reverse)

        for field in self._meta.local_many_to_many:
            # from IPython import embed;embed()
            if (
                field.remote_field.through
                and field.remote_field.through._meta.auto_created
                and field.name not in self.clone_related
            ):
                man = getattr(self, field.name)
                man.clear()

    def _delete_reverse(self, reverse):
        cache = self._get_field_map()
        rel = cache[reverse]
        m2m = rel.is_relation and rel.many_to_many
        direct = not rel.auto_created or rel.concrete

        if m2m:
            if direct:
                manager = getattr(self, reverse)
            else:
                assert False, "cloning reverse m2m is not currently supported"
        else:
            manager = getattr(self, rel.get_accessor_name())

        # Have to do this the slow way so all possible
        # objects are deleted
        for obj in manager.all():
            obj.delete()

    def delete(self, *args, **kwargs):
        """
        Delete clonable relations first, since they may be
        objects that wouldn't otherwise be deleted.

        Calls super to actually delete the object.
        """
        skip_reverses = kwargs.pop("skip_reverses", False)
        if not skip_reverses:
            self._delete_reverses()

        return super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Takes an optional last_save keyword argument
        other wise last_save will be set to timezone.now()

        Calls super to actually save the object.
        """
        self.last_save = kwargs.pop("last_save", timezone.now())
        super().save(*args, **kwargs)

    @property
    def clone_related(self):
        return getattr(self, "_clone_related", [])

    @classmethod
    def register_related(cls, related_name):
        """
        Register a related item that should be cloned
        when this model is.

        :param related_name: Use the name you would use in filtering
            i.e.: book not book_set.
        """

        if not hasattr(cls, "_clone_related"):
            cls._clone_related = []

        if not isinstance(cls._clone_related, list):
            cls._clone_related = list(cls._clone_related)

        if related_name not in cls._clone_related:
            cls._clone_related.append(related_name)

    class Meta:
        abstract = True


def get_meta(meta, **kwargs):
    new_meta = type("Meta", (object,), {})
    if meta:
        for k, v in list(meta.__dict__.items()):
            # Python3: `meta.__dict__` is <class 'mappingproxy'>
            if v is not None and (k != "__dict__" and k != "__weakref__"):
                setattr(new_meta, k, v)

    for k, v in list(kwargs.items()):
        if v is not None:
            setattr(new_meta, k, v)
    return new_meta


def add_managers(attrs):
    if "objects" not in attrs and "_default_manager" not in attrs:
        attrs["objects"] = manager.VersionManager()
    attrs["normal"] = models.Manager()
    attrs["_base_manager"] = models.Manager()
    return attrs


class SharedMeta(models.base.ModelBase):
    def add_to_class(cls, name, value):
        if not cls._meta.abstract:
            version_model = cls.version_model
            base_model = cls.base_model

            if isinstance(value, models.ManyToManyField) or isinstance(
                value, models.ForeignKey
            ):
                if version_model:
                    if isinstance(value, models.ManyToManyField):
                        from .fields import M2MFromVersion

                        if not isinstance(value, M2MFromVersion):
                            raise TypeError(
                                f"The model {name} cannot contain a ManyToManyField "
                                "use M2MFromVersion instead"
                            )
                        value.db_table = version_model._meta.get_field(
                            name
                        ).m2m_db_table()
                else:
                    if (
                        value.remote_field.related_name
                        and not value.remote_field.related_name.endswith("+")
                    ):
                        value.remote_field.related_name = (
                            f"{value.remote_field.related_name}_version"
                        )

                # relationships to self should always point
                # to the base model
                if value.remote_field.model == related.RECURSIVE_RELATIONSHIP_CONSTANT:
                    value.remote_field.model = base_model

        # Django 1.10 doesn't allow us to modify `_base_manager` attr:
        # https://docs.djangoproject.com/en/1.10/topics/db/managers/#django.db.models.Model._base_manager
        if name == "_base_manager":
            if not value.name:
                value.name = name
            value.model = cls
            setattr(cls._meta, "base_manager_name", "normal")
        else:
            super().add_to_class(name, value)


class VersionViewMeta(SharedMeta):
    """
    Meta class for models that are implementing the
    database view to hide the fact there are two
    underlying models and fetch version with a parent
    automatically.
    """

    def __new__(cls, name, bases, attrs):
        meta = attrs.pop("Meta", None)

        # If this is an abstract model don't do anything special
        if meta and getattr(meta, "abstract", False):
            attrs["Meta"] = meta
            return super().__new__(cls, name, bases, attrs)
        # List of non field attrs we want to copy
        copy_attrs = attrs.pop("_copy_extra_attrs", [])

        # Construct the base class.

        # An abstract version of the base model
        # needed to make the view model inherit from it.
        ab_base_model = BaseModel

        base_model = None
        base_name = None

        # If we have a custom base model use it
        if meta and hasattr(meta, "_base_model"):
            base_model = meta._base_model
            meta._base_model = None
            assert issubclass(base_model, BaseModel)

            base_name = f"{base_model._meta.app_label}.{base_model.__name__}"

            # Create a new abstract base model we can inherit from
            ab_base_model = type(
                str(base_name) + "Abs",
                (BaseModel,),
                {
                    "__module__": base_model.__module__,
                    "Meta": get_meta(None, abstract=True),
                },
            )

            # We only copy local fields on purpose
            # No relations allowed on base models
            parent_fields = base_model._meta.local_fields
            for field in parent_fields:
                if not field.primary_key:
                    try:
                        ab_base_model._meta.get_field(field.name)
                    except FieldDoesNotExist:
                        ab_base_model.add_to_class(field.name, copy.deepcopy(field))

        # Create a base model that isn't abstract
        else:
            base_name = f"{name}_base"
            base_abstract = meta and getattr(meta, "abstract", False) or False
            base_model = type(
                base_name,
                (BaseModel,),
                {
                    "__module__": attrs.get("__module__"),
                    "Meta": get_meta(None, abstract=base_abstract),
                },
            )

        versioned_attrs = OrderedDict()
        versioned_attrs["__module__"] = attrs.get("__module__")
        versioned_attrs["Meta"] = get_meta(meta, managed=True)

        # Create the version primary key
        versioned_attrs["vid"] = models.AutoField(primary_key=True)

        # Create a FK to base
        versioned_attrs["object"] = models.ForeignKey(
            base_name,
            related_name=base_model.get_related_name(name),
            on_delete=models.CASCADE,
        )

        versioned_attrs = add_managers(versioned_attrs)

        # Copy all the fields
        for k, v in list(attrs.items()):
            if k in copy_attrs:
                versioned_attrs[k] = copy.deepcopy(v)
            elif isinstance(v, Field):
                v = copy.deepcopy(v)
                versioned_attrs[k] = v

        version_bases = [BaseVersionedModel]
        custom_version = False
        for base in bases:
            if issubclass(base, models.Model) and not base._meta.abstract:
                raise TypeError(
                    f"You cannot use a non abstract base model {base!r} "
                    f"with a VersionView model {name}."
                )

            if issubclass(base, BaseModel):
                raise TypeError(
                    f"VersionView model {cls!r} can't inherit from a BaseModel {base!r}"
                )

            if issubclass(base, BaseVersionedModel):
                if not issubclass(type(base), VersionViewMeta):
                    if custom_version:
                        raise TypeError(
                            "Django inheritence will only allow VersionView model "
                            f"{cls!r} to inherit from one and only one BaseVersionedModel"
                        )
                    else:
                        version_bases[0] = base
                        custom_version = True
            else:
                version_bases.append(base)

        vname = f"{name}VersionReferences"
        v_mod = type(
            vname,
            (object,),
            {
                "__module__": attrs.get("__module__"),
                "base_model": base_model,
                "version_model": None,
            },
        )
        version_bases.append(v_mod)
        setattr(sys.modules[attrs.get("__module__")], vname, v_mod)
        version_model = SharedMeta(
            f"{name}_version", tuple(version_bases), versioned_attrs
        )

        # Make sure the managed = False
        attrs["Meta"] = get_meta(meta, managed=False)

        attrs = add_managers(attrs)

        new_bases = [ab_base_model]
        for base in bases:
            if not issubclass(base, BaseModel):
                new_bases.append(base)

        rname = f"{name}References"
        ref_mod = type(
            rname,
            (object,),
            {
                "__module__": attrs.get("__module__"),
                "base_model": base_model,
                "version_model": version_model,
            },
        )
        setattr(sys.modules[attrs.get("__module__")], rname, ref_mod)
        new_bases.append(ref_mod)

        mod = super().__new__(cls, name, tuple(new_bases), attrs)

        # All the models should know about each other
        base_model._meta._view_model = mod
        base_model._meta._version_model = version_model

        version_model._meta._view_model = mod
        version_model._meta._base_model = base_model

        mod._meta._version_model = version_model
        mod._meta._base_model = base_model
        mod._meta._is_view = True

        # Register signals
        published_delete_signal.connect(
            mod.handle_published_delete_signal,
            sender=mod,
            dispatch_uid=mod.__name__ + "delete",
        )
        published_signal.connect(
            mod.handle_published_signal,
            sender=mod,
            dispatch_uid=mod.__name__ + "publish",
        )
        models.signals.pre_delete.connect(
            mod.handle_pre_delete_signal,
            sender=mod,
            dispatch_uid=mod.__name__ + "pre_delete",
        )
        models.signals.post_delete.connect(
            mod.handle_post_delete_signal,
            sender=mod,
            dispatch_uid=mod.__name__ + "post_delete",
        )

        return mod

    def add_to_class(cls, name, value):
        try:
            BaseVersionedModel._meta.get_field(name)
            cls._meta.get_field(name)
            return
        except FieldDoesNotExist:
            pass

        return super().add_to_class(name, value)


class VersionModelMeta(models.base.ModelBase):
    """
    Meta class for models that are transparent
    about there being two models and that fetch
    versions/parents manually.
    """

    def add_to_class(cls, name, value):
        """
        Django 1.10 doesn't allow us to modify `_base_manager` attr:
        https://docs.djangoproject.com/en/1.10/topics/db/managers/#django.db.models.Model._base_manager
        """
        if name == "_base_manager":
            if not value.name:
                value.name = name
            value.model = cls
            setattr(cls._meta, "base_manager_name", "normal")
        else:
            super().add_to_class(name, value)

    def __new__(cls, name, bases, attrs):
        meta = attrs.get("Meta", None)
        base_model = None

        # Only do this work if we have a non-abstract model
        if not getattr(meta, "abstract", False):
            if not getattr(meta, "_base_model", None) or not issubclass(
                meta._base_model, BaseModel
            ):
                raise TypeError(
                    f"{name} must declare a subclass of BaseModel as "
                    "it's _base_model in it's meta class"
                )
            else:
                base_model = meta._base_model
                meta._base_model = None

            # Create a FK to base
            attrs["object"] = models.ForeignKey(
                base_model.__name__,
                related_name=base_model.get_related_name(name),
                editable=False,
                on_delete=models.CASCADE,
            )
            attrs = add_managers(attrs)

        mod = super().__new__(cls, name, bases, attrs)

        # The models should know about each other
        if base_model:
            mod._meta._base_model = base_model
            base_model._meta._version_model = mod

        return mod


class BaseModel(models.Model):

    """
    Abstract model for the base model of your versioned model.
    This is the model where there is only one row per
    item and that each version relates to. Any custom base
    models must inherit from this one.

    Provides the following fields:

    `is_published`: Does this item contain any live versions?

    `created_date`: Date this item was originally created.

    `v_last_save`: When the current live item was last saved,
    may be null.
    """

    is_published = models.BooleanField(editable=False, default=False)

    created_date = models.DateTimeField(default=timezone.now, editable=False)

    v_last_save = models.DateTimeField(null=True, editable=False)

    _version = None

    class Meta:
        abstract = True

    @classmethod
    def get_related_name(cls, name=None):
        """
        Override this method and provide a unique related_name to avoid
        conflicts if multiple VersionViews are using the same _base_model
        in their Meta attributes.
        """
        return "version"

    def get_version(self, state=None, date=None):
        """
        Get a particular version of an item

        :param state: The state you want to get.
        :param date: Get a version that was published before or on this date.
        """

        version_model = self._meta._version_model
        q = version_model.objects.filter(object_id=self.pk)
        if state:
            q = version_model.normal.filter(object_id=self.pk, state=state)

        if date:
            q = q.filter(date_published__lte=date)

        q = q.order_by("-date_published")

        results = q[:1]
        if results:
            return results[0]
        return None

    def _get_version(self):
        if self._version is None:
            self._version = self.get_version()
        return self._version

    def _set_version(self, version):
        self._version = version

    current_version = property(_get_version, _set_version)


class BaseVersionedModel(Cloneable, Schedulable):
    """
    Base abstract model for the versioned model data.

    Inherits from both Cloneable and Schedulable.

    Implemented by both VersionView and VersionModel.
    """

    # How many archived versions should we keep?
    NUM_KEEP_ARCHIVED = 5

    # Supported States
    PUBLISHED = "published"
    DRAFT = "draft"
    SCHEDULED = "scheduled"
    ARCHIVED = "archived"

    UNIQUE_STATES = (PUBLISHED, DRAFT)

    STATES = (
        (PUBLISHED, PUBLISHED),
        (SCHEDULED, SCHEDULED),
        (DRAFT, DRAFT),
        (ARCHIVED, ARCHIVED),
    )

    state = models.CharField(max_length=50, choices=STATES, editable=False)

    # Date the item was last scheduled to go live
    last_scheduled = models.DateTimeField(null=True, editable=False)

    # The date the item is supposed-to/actually went live
    date_published = models.DateTimeField(null=True, editable=False)

    # username who scheduled the publish
    user_published = models.CharField(max_length=255, null=True, editable=False)

    class Meta:
        abstract = True

    def prep_for_clone(self):
        self.vid = None
        self.user_published = None

    def unpublish(self):
        """
        Unpublish this item.

        This will set and currently published versions to
        the archived state and delete all currently scheduled
        versions.
        """
        assert self.state == self.DRAFT

        with xact():
            self._publish(published=False)

            # Delete all scheduled items
            klass = self.get_version_class()
            for obj in klass.normal.filter(
                object_id=self.object_id, state=self.SCHEDULED
            ):
                obj.delete()

    def publish(self, user=None, when=None):
        """
        Publishes a item and any sub items.
        A new transaction will be started if
        we aren't already in a transaction.

        Should only be run on draft items
        """

        assert self.state == self.DRAFT

        user_published = "code"
        if user:
            user_published = user.username

        now = timezone.now()
        state = self.PUBLISHED if not when or when <= now else self.SCHEDULED

        with xact():
            # If this item hasn't gone live yet delete the old scheduled items
            published = False
            if getattr(self._meta, "_is_view", False):
                published = self.is_published
            else:
                published = self.object.is_published

            if not when and not published and self.last_scheduled:
                klass = self.get_version_class()
                for obj in klass.normal.filter(
                    object_id=self.object_id,
                    last_scheduled=self.last_scheduled,
                    state=self.SCHEDULED,
                ):
                    obj.delete()

            # Drafts get preserved so save the
            # time we last cloned this
            if self.state == self.DRAFT:
                self.last_scheduled = now
                self.date_published = when or now
                self.save(last_save=now)

            self._clone()

            self.user_published = user_published
            self.state = state
            self.save()

            if state == self.PUBLISHED:
                self._publish(date_published=when)
            else:
                self.schedule(when=when)

    def _delete_reverse(self, reverse):
        try:
            return super()._delete_reverse(reverse)
        except KeyError:
            # Could this be on the view model?
            if not hasattr(self._meta, "_view_model"):
                raise

            # XXX: These query CAN'T do any joins with a
            # version view table, or bad things will happen
            # need to test for that
            sub = self._meta._view_model(object_id=self.object_id, vid=self.vid)
            return sub._delete_reverse(reverse)

    def _gather_reverse(self, reverse):
        try:
            return super()._gather_reverse(reverse)
        except KeyError:
            # Could this be on the view model?
            if not hasattr(self._meta, "_view_model"):
                raise

            # These query CAN'T do any joins with a
            # version view table, or bad things will happen
            # need to test for that
            sub = self._meta._view_model(object_id=self.object_id, vid=self.vid)
            return sub._gather_reverse(reverse)

    @property
    def clone_related(self):
        if hasattr(self._meta, "_view_model"):
            return getattr(self._meta._view_model, "_clone_related", [])
        return getattr(self, "_clone_related", [])

    def make_draft(self):
        """
        Make this version the draft
        """
        assert self.__class__ == self.get_version_class()

        # If this is draft do nothing
        if self.state == self.DRAFT:
            return

        with xact():
            # Delete whatever is currently this draft
            try:
                klass = self.get_version_class()
                old_draft = klass.normal.get(object_id=self.object_id, state=self.DRAFT)
                old_draft.delete()
            except klass.DoesNotExist:
                pass

            # Set this to draft and save
            self.state = self.DRAFT
            # Make last_scheduled and last save match on draft
            self.last_save = self.last_scheduled
            self._clone()

    def get_version_class(self):
        klass = self.__class__
        if getattr(self._meta, "_is_view", False):
            klass = self._meta._version_model
        return klass

    def _publish(self, published=True, **kwargs):
        with xact():
            filter_args = {
                "state": self.PUBLISHED,
                "object_id": self.object_id,
            }

            klass = self.get_version_class()
            klass.normal.filter(**filter_args).update(state=self.ARCHIVED)

            now = timezone.now()

            if published:
                self.state = self.PUBLISHED
                self.date_published = kwargs.get("date_published") or now
            else:
                self.last_scheduled = None
                self.date_published = None

            self.save()

            # Update the base model as published and cache the scheduled
            # date for comparisons.
            self._meta._base_model.objects.filter(pk=self.object_id).update(
                is_published=published, v_last_save=self.last_scheduled
            )
            self.is_published = published
            self.v_last_save = self.last_scheduled

        # Send published signal
        klass = self.__class__
        if hasattr(self._meta, "_view_model"):
            klass = self.__class__

        published_signal.send(sender=klass, instance=self)

    def purge_archives(self):
        """
        Delete older archived items.

        Use the class attribute NUM_KEEP_ARCHIVED to control
        how many items are kept.
        """

        klass = self.get_version_class()
        qs = klass.normal.filter(
            object_id=self.object_id, state=self.ARCHIVED
        ).order_by("-last_save")[self.NUM_KEEP_ARCHIVED :]

        for obj in qs:
            obj._delete_reverses()
            klass.normal.filter(vid=obj.vid).delete()

    def status_line(self):
        """
        Returns a status line for an item.

        Only really interesting when called for a draft
        item as it can tell you if the draft is the same as
        another version.
        """

        date = self.date_published
        status = self.state.title()
        if self.state == self.DRAFT:
            # Check if this item has changed since
            # our last publish
            status = "Draft saved"
            date = self.last_save
            if date and self.last_save == self.last_scheduled:
                # We need to figure out if the item it is based on
                # is either live now or will be live at some point.

                # If last_scheduled is less than or equal to
                # v_last_save this item is or will go live
                # at some point. Otherwise it won't
                # so we'll leave state as draft.
                if self.v_last_save:
                    if self.last_scheduled >= self.v_last_save:
                        status = self.PUBLISHED.title()

                    # The date this was scheduled is greater than
                    # what is currently live, this will go live at
                    # some point
                    if self.last_scheduled > self.v_last_save:
                        status = "Publish Scheduled"
                else:
                    status = "Publish Scheduled"

                date = self.date_published

        if date:
            status = f"{status}: {formats.date_format(date, 'SHORT_DATE_FORMAT')}"
        return status

    def schedule(self, when=None, action=None, **kwargs):
        """
        Schedule this item to be published.

        :param when: Date/time when this item should go live. None means now.
        """
        action = "_publish"
        super().schedule(when=when, action=action, **kwargs)


class VersionView(BaseVersionedModel, metaclass=VersionViewMeta):
    """
    Abstract base model for implementations
    that are pointing to a database view
    that handles the joining of the two tables for you.

    Implementations can then rely on schema path switching
    to filter on either draft or published states.

    Outside of that you will either need to query the underlying
    model or use the public schema and specify the correct
    filter parameters to filter out the states/versions that you
    are not interested in wherever this model is used in a query.

    Inherits from `BaseVersionedModel`

    Contains the following fields:

    vid: The unique id of this version, this is unique within this table.

    object_id: The id of the item on the base_model.

    last_scheduled: Date the item was last scheduled to go live. May be null.

    date_published: The date the item is supposed-to/actually went live.
    May be null.

    user_published: The username of the user who published. May be null.

    last_save: The date this version was last_saved.
    """

    # What version id is this
    vid = models.PositiveIntegerField(unique=True, editable=False)

    # What the object id, same as pk just for consistency
    object_id = models.PositiveIntegerField(editable=False)

    class Meta:
        abstract = True

    def get_scheduled_filter_args(self):
        return {"vid": self.vid}

    def _clone(self, *args, **kwargs):
        # We should only clone from a draft view in that mode
        assert self.state == self.DRAFT
        return super()._clone(*args, **kwargs)

    def delete(self, **kwargs):
        with xact():
            # Delete's happen in default schema so that all
            # versions are found.
            with manager.SwitchSchema("public"):
                super().delete(**kwargs)

    def _get_should_save_base(self):
        if not hasattr(self, "_should_save_base"):
            self._should_save_base = False
        return self._should_save_base

    def _set_should_save_base(self, value):
        self._should_save_base = value

    should_save_base = property(_get_should_save_base, _set_should_save_base)

    def save(self, *args, **kwargs):
        models.signals.pre_save.send(
            sender=self.__class__,
            instance=self,
            raw=kwargs.get("raw"),
            using=kwargs.get("using"),
        )

        last_save = kwargs.pop("last_save", timezone.now())
        with xact():
            # Make sure we have a state
            if not self.vid:
                self.state = self.DRAFT

            # Get a version instance
            version = self._meta._version_model()
            base = self._meta._base_model()

            # Build out the version
            for f in self._meta.local_fields:
                try:
                    attname = f.name
                    version._meta.get_field(attname)
                    if isinstance(f, models.ForeignKey):
                        attname = f"{f.name}_id"

                    setattr(version, attname, getattr(self, attname))

                # version doesn't have this field
                except FieldDoesNotExist:
                    if self.should_save_base:
                        setattr(base, f.name, getattr(self, attname))

            # Make sure we have a base
            if not self.object_id or self.should_save_base:
                base.is_published = self.is_published
                base.pk = self.pk

                base.save(*args, **kwargs)
                self.pk = base.pk

                # set base object
                version.object = base
            else:
                version.object_id = self.object_id

            # Save the version and copy the db state
            version.save(last_save=last_save)

            self.vid = version.vid
            self._state = version._state
            self.state = version.state
            self.last_save = version.last_save
        models.signals.post_save.send(
            sender=self.__class__,
            instance=self,
            raw=kwargs.get("raw"),
            using=kwargs.get("using"),
        )

    def validate_unique(self, *args, **kwargs):
        """
        Calls super validate_unique and, after that is done, runs
        through any field_names listed in `self.versioned_unique`
        and checks that this is the only item with this name.

        This relies on using the public schema when running this
        check.

        These checks contain race conditions since it all happens
        before saves and so should not be relied upon for uniqueness
        but help with form validation.
        """

        super().validate_unique(*args, **kwargs)
        if hasattr(self, "versioned_unique"):
            unique_checks = []
            for field in self.versioned_unique:
                unique_checks.append((self.__class__, (field,)))
            errors = self._perform_unique_checks(unique_checks)

            if errors:
                raise ValidationError(errors)

    @classmethod
    def register_related(cls, related_name):
        super().register_related(related_name)
        cls._meta._version_model.register_related(related_name)

    @classmethod
    def invalidate_cache(cls, sender, instance, **kwargs):
        try:
            try:
                from ..cache import cache_manager
            except ValueError:
                from cache import cache_manager
            cache_manager.invalidate_cache(cls, instance=instance)
        except ImportError:
            pass

    @classmethod
    def handle_published_delete_signal(cls, sender, instance, **kwargs):
        cls.invalidate_cache(sender, instance, **kwargs)

    @classmethod
    def handle_published_signal(cls, sender, instance, **kwargs):
        cls.invalidate_cache(sender, instance, **kwargs)

    @classmethod
    def handle_pre_delete_signal(cls, sender, instance, **kwargs):
        with xact():
            klass = cls._meta._version_model
            for obj in klass.normal.filter(object_id=instance.object_id):
                obj._delete_reverses()

    @classmethod
    def handle_post_delete_signal(cls, sender, instance, **kwargs):
        published_delete_signal.send(sender, instance=instance)


class VersionModel(BaseVersionedModel, metaclass=VersionModelMeta):
    """
    Abstract base model for implementations
    that are transparent about the fact that
    there are two models/tables for this model.

    Implementations that use this model will need to take
    care to specify the correct filter parameters to filter
    out the states/versions that you are not interested in
    wherever this model is used in a query. This may not
    always be obvious especially in complex joins or when
    using select_related or prefetch_related.

    Inherits from `BaseVersionedModel`

    Contains the following fields:

    vid: The unique id of this version, this is unique within this table.

    object: A foreign key to the base item model.

    last_scheduled: Date the item was last scheduled to go live. May be null.

    date_published: The date the item is supposed-to/actually went live.
    May be null.

    user_published: The username of the user who published. May be null.

    last_save: The date this version was last_saved.
    """

    # Setup the primary key
    vid = models.AutoField(primary_key=True)

    class Meta:
        abstract = True

    def _get_v_last_save(self):
        return self.object.v_last_save

    def _set_v_last_save(self, value):
        return None

    v_last_save = property(_get_v_last_save, _set_v_last_save)

    def validate_unique(self, *args, **kwargs):
        """
        Calls super validate_unique and, after that is done, runs
        through any field_names listed in `self.versioned_unique`
        and checks that this is the only item with this name.

        These checks contain race conditions since it all happens
        before saves and so should not be relied upon for uniqueness
        but help with form validation.
        """

        super(BaseVersionedModel, self).validate_unique(*args, **kwargs)
        if hasattr(self, "versioned_unique"):
            errors = {}

            for field in self.versioned_unique:
                lookup_kwargs = {field: getattr(self, field)}
                qs = self.__class__._default_manager.filter(**lookup_kwargs)

                # Exclude the current object from the query if we are editing
                # an instance (as opposed to creating a new one)
                if self.object_id:
                    qs = qs.exclude(object_id=self.object_id)

                if qs.exists():
                    msg = self.unique_error_message(self.__class__, (field,))
                    errors[field] = msg

            if errors:
                raise ValidationError(errors)

    def save(self, *args, **kwargs):
        """
        Saves this item.

        Creates a default base if there isn't
        one already.
        """
        with xact():
            if not self.vid:
                self.state = self.DRAFT

                if not self.object_id:
                    base = self._meta._base_model(is_published=False)
                    base.save(*args, **kwargs)
                    self.object = base

            super().save(*args, **kwargs)


# Setup signal for published
published_signal = dispatch.Signal(providing_args=["instance"])
published_delete_signal = dispatch.Signal(providing_args=["instance"])



from .. import Provider as GeoProvider


class Provider(GeoProvider):

    def local_latitude(self):
        return self.coordinate(center=47.60707, radius=1)

    def local_longitude(self):
        return self.coordinate(center=13.37208, radius=2)



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("taggit", "0001_initial")]

    operations = [
        migrations.AlterIndexTogether(
            name="taggeditem", index_together={("content_type", "object_id")}
        )
    ]



import six


@six.python_2_unicode_compatible
class LiveServer(object):
    """The liveserver fixture

    This is the object that the ``live_server`` fixture returns.
    The ``live_server`` fixture handles creation and stopping.
    """

    def __init__(self, addr):
        import django
        from django.db import connections
        from django.test.testcases import LiveServerThread
        from django.test.utils import modify_settings

        connections_override = {}
        for conn in connections.all():
            # If using in-memory sqlite databases, pass the connections to
            # the server thread.
            if (
                conn.settings_dict["ENGINE"] == "django.db.backends.sqlite3"
                and conn.settings_dict["NAME"] == ":memory:"
            ):
                # Explicitly enable thread-shareability for this connection
                conn.allow_thread_sharing = True
                connections_override[conn.alias] = conn

        liveserver_kwargs = {"connections_override": connections_override}
        from django.conf import settings

        if "django.contrib.staticfiles" in settings.INSTALLED_APPS:
            from django.contrib.staticfiles.handlers import StaticFilesHandler

            liveserver_kwargs["static_handler"] = StaticFilesHandler
        else:
            from django.test.testcases import _StaticFilesHandler

            liveserver_kwargs["static_handler"] = _StaticFilesHandler

        if django.VERSION < (1, 11):
            host, possible_ports = parse_addr(addr)
            self.thread = LiveServerThread(host, possible_ports, **liveserver_kwargs)
        else:
            try:
                host, port = addr.split(":")
            except ValueError:
                host = addr
            else:
                liveserver_kwargs["port"] = int(port)
            self.thread = LiveServerThread(host, **liveserver_kwargs)

        self._live_server_modified_settings = modify_settings(
            ALLOWED_HOSTS={"append": host}
        )

        self.thread.daemon = True
        self.thread.start()
        self.thread.is_ready.wait()

        if self.thread.error:
            raise self.thread.error

    def stop(self):
        """Stop the server"""
        self.thread.terminate()
        self.thread.join()

    @property
    def url(self):
        return "http://%s:%s" % (self.thread.host, self.thread.port)

    def __str__(self):
        return self.url

    def __add__(self, other):
        return "%s%s" % (self, other)

    def __repr__(self):
        return "<LiveServer listening at %s>" % self.url


def parse_addr(specified_address):
    """Parse the --liveserver argument into a host/IP address and port range"""
    # This code is based on
    # django.test.testcases.LiveServerTestCase.setUpClass

    # The specified ports may be of the form '8000-8010,8080,9200-9300'
    # i.e. a comma-separated list of ports or ranges of ports, so we break
    # it down into a detailed list of all possible ports.
    possible_ports = []
    try:
        host, port_ranges = specified_address.split(":")
        for port_range in port_ranges.split(","):
            # A port range can be of either form: '8000' or '8000-8010'.
            extremes = list(map(int, port_range.split("-")))
            assert len(extremes) in (1, 2)
            if len(extremes) == 1:
                # Port range of the form '8000'
                possible_ports.append(extremes[0])
            else:
                # Port range of the form '8000-8010'
                for port in range(extremes[0], extremes[1] + 1):
                    possible_ports.append(port)
    except Exception:
        raise Exception('Invalid address ("%s") for live server.' % specified_address)

    return host, possible_ports



from django import http
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import ProtectedError
from django.utils.encoding import force_text
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.list import MultipleObjectMixin

from . import renders, transaction
from .base_views import ModelCMSMixin, ModelCMSView
from .forms import WhenForm
from .models import CMSLog

CHECKBOX_NAME = "_selected"


class ActionView(ModelCMSMixin, MultipleObjectMixin, ModelCMSView):
    """
    Base class for defining actions that can be executed on one \
    or many objects. An ActionView subclass that will be used as a 'mass \
    action' should be registered with the Bundle in the Meta class. \
    If it is going to be used as a single-item action, it should be \
    registered as an item_view in the same way. If an ActionView will be \
    used as both a mass action and a single-item action, it can simply be \
    registered as an action view.

        i.e. (in the Bundle)
            class Meta:
                action_views = ['action1', 'action2']

    The actions that appear in a specific ListView class can be specified \
    with the `action_view` kwarg. Any action that appears here should first\
    be registered with the bundle. Views will, by default, show all actions \
    that are registered to the bundle.

    If ListView formsets are also being used, selecting an action will \
    override any edits for the formset. A formset will still process \
    correctly if no action is selected.

    :param default_template: defaults to cms/action_confirmation.html.
    :param short_description: Description of action to display. \
    Defaults to the name of the view.
    :param redirect_to_view: Defaults to 'main'
    :param confirmation_message: Message that the intermediate \
    confirmation page will display to the user before action \
    is executed on multiple items.
    :param confirmation_message_single: Message that the intermediate \
    confirmation page will display to the user before action \
    is executed on a single item.
    """

    short_description = None
    redirect_to_view = "main_list"
    confirmation_message = (
        "Please confirm that you want to {action_name} the following {bundle_name}:"
    )
    confirmation_message_single = (
        "Please confirm that you want to {action_name} the {bundle_name}"
    )
    default_template = "cms/action_confirmation.html"

    object = None
    action_name = None
    show_navigation = False

    def render(self, *args, **kwargs):
        if "action" not in kwargs:
            if not self.action_name:
                kwargs["action"] = "Yes"
            else:
                kwargs["action"] = self.action_name
        return super().render(*args, **kwargs)

    def get_navigation(self):
        nav = None
        if self.show_navigation:
            if not self.object and self.name not in self.bundle._meta.action_views:
                if self.bundle.parent:
                    nav = self.bundle.parent.get_navigation(self.request, **self.kwargs)
            else:
                nav = super().get_navigation()
        return nav

    def process_action(self, request, queryset):
        """
        Can be overriden to define the actions performed
        on the given queryset.
        Arguments are original request and queryset of
        objects to be modified by the action.

        :param request: Original HTTP request
        :param queryset: Queryset of objects to modify

        Returns render response. By default, a render redirect
        will be returned to the view specified by `redirect_to_view.`
        """
        pass

    def get_confirmation_message(self, queryset):
        if len(queryset) == 1:
            confirmation_msg = self.confirmation_message_single.format(
                action_name=force_text(self.action_name).lower(),
                bundle_name=force_text(self.bundle.get_single_title()).lower(),
            )
        else:
            confirmation_msg = self.confirmation_message.format(
                action_name=force_text(self.action_name).lower(),
                bundle_name=force_text(self.bundle.get_title()).lower(),
            )
        return confirmation_msg

    def get_context_data(self, **kwargs):
        """
        Hook for adding arguments to the context.
        """

        context = {"obj": self.object}
        if "queryset" in kwargs:
            context["conf_msg"] = self.get_confirmation_message(kwargs["queryset"])
        context.update(kwargs)
        return context

    def get_object(self):
        """
        If a single object has been requested, will set
        `self.object` and return the object.
        """
        slug = self.kwargs.get(self.slug_url_kwarg, None)

        if slug is not None:
            queryset = self.get_queryset()
            slug_field = self.slug_field
            queryset = queryset.filter(**{slug_field: slug})
            try:
                self.object = queryset.get()
            except ObjectDoesNotExist:
                raise http.Http404
        return self.object

    def write_message(self, status=messages.INFO, message=None):
        """
        Same behavior as parent, except defaults to the message
        "Saved" if no message is present.
        """
        if not message:
            message = "Saved."
        messages.add_message(self.request, status, message)
        return message

    def get_done_url(self):
        """
        Returns the url to redirect to after a successful update.
        The get_view_url will be called on the current bundle using
        `self.redirect_to_view` as the view name.
        """
        data = dict(self.kwargs)
        data.pop(self.slug_url_kwarg, None)
        url = self.bundle.get_view_url(self.redirect_to_view, self.request.user, data)
        return self.customized_return_url(url)

    def get_selected(self, request):
        """
        Returns a queryset of the selected objects as specified by \
        a GET or POST request.
        """
        obj = self.get_object()
        # if single-object URL not used, check for selected objects
        if not obj:
            if request.GET.get(CHECKBOX_NAME):
                selected = request.GET.get(CHECKBOX_NAME).split(",")
            else:
                selected = request.POST.getlist(CHECKBOX_NAME)
        else:
            selected = [obj.pk]

        queryset = self.get_queryset().filter(pk__in=selected)
        return queryset

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests.
        Calls the `render` method with the following
        items in context:

        * **queryset** - Objects to perform action on
        """

        queryset = self.get_selected(request)
        return self.render(request, queryset=queryset)

    def post(self, request, *args, **kwargs):
        """
        Method for handling POST requests.
        Checks for a modify confirmation and performs
        the action by calling `process_action`.

        """
        queryset = self.get_selected(request)

        if request.POST.get("modify"):
            response = self.process_action(request, queryset)
            if not response:
                url = self.get_done_url()
                return self.render(request, redirect_url=url)
            else:
                return response
        else:
            return self.render(request, redirect_url=request.build_absolute_uri())

    def as_string(self, *args, **kwargs):
        if self.render_type == "option":
            return self.short_description

        return super().as_string(*args, **kwargs)


class DeleteActionView(ActionView):
    """
    View for deleting one or more objects.
    Inherits from ActionView.
    Used through out the CMS as the default object_header view so
    this view by default adds an `object_header` renderer that
    uses the `object_header_tmpl` template.

    :param redirect_to_view: Defaults to 'main_list'.
    """

    short_description = "Delete selected items"
    action_name = "Delete"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.renders["object_header"] = renders.RenderString(
            template=self.object_header_tmpl
        )

    def process_action(self, request, queryset):
        """
        Deletes the object(s). Successful deletes are logged.
        Returns a 'render redirect' to the result of the
        `get_done_url` method.

        If a ProtectedError is raised, the `render` method
        is called with message explaining the error added
        to the context as `protected`.
        """
        count = 0
        try:
            with transaction.commit_on_success():
                for obj in queryset:
                    self.log_action(obj, CMSLog.DELETE)
                    count += 1
                    obj.delete()
            msg = f"{count} object{'' if count == 1 else 's'} deleted."
            url = self.get_done_url()
            return self.render(request, redirect_url=url, message=msg)
        except ProtectedError as e:
            protected = []
            for x in e.protected_objects:
                if hasattr(x, "delete_blocked_message"):
                    protected.append(x.delete_blocked_message())
                else:
                    protected.append(f"{x._meta.verbose_name} - {x}")
            msg = "Cannot delete some objects because the following objects depend on them:"
            return self.render(request, error_msg=msg, errors=protected)


class PublishActionView(ActionView):
    """
    View for publishing an object. Inherits from
    ModelCMSMixin, SingleObjectMixin, ModelCMSView.
    Assumes the given model is versionable.

    :param default_template: Defaults to cms/publish_action.html.
    :param short_description: Defaults to 'Publish selected items'
    """

    default_template = "cms/publish_action.html"
    short_description = "Publish selected items"
    form = WhenForm
    action_name = "Publish"

    def get_object(self):
        """
        Get the object for publishing
        Raises a http404 error if the object is not found.
        """
        obj = super().get_object()

        if obj:
            if not hasattr(obj, "publish"):
                raise http.Http404

        return obj

    def get_object_url(self, obj):
        """
        Returns the url to link to the object
        The get_view_url will be called on the current bundle using
        'edit` as the view name.
        """
        return self.bundle.get_view_url(
            "edit", self.request.user, {"object": obj}, self.kwargs
        )

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests. Passes the
        following arguments to the context:

        * **queryset** - The queryset of objects to publish.
        * **publish_form** - An instance of WhenForm.
        """

        queryset = self.get_selected(request)
        return self.render(request, queryset=queryset, publish_form=self.form())

    def process_action(self, request, queryset):
        """
        Publishes the selected objects by passing the value of \
        'when' to the object's publish method. The object's \
        `purge_archives` method is also called to limit the number \
        of old items that we keep around. The action is logged as \
        either 'published' or 'scheduled' depending on the value of \
        'when', and the user is notified with a message.

        Returns a 'render redirect' to the result of the \
        `get_done_url` method.
        """
        form = self.form(request.POST)
        if form.is_valid():
            when = form.cleaned_data.get("when")
            count = 0
            for obj in queryset:
                count += 1
                obj.publish(user=request.user, when=when)
                obj.purge_archives()
                object_url = self.get_object_url(obj)
                if obj.state == obj.PUBLISHED:
                    self.log_action(obj, CMSLog.PUBLISH, url=object_url)
                else:
                    self.log_action(obj, CMSLog.SCHEDULE, url=object_url)
            message = f"{count} object{'' if count == 1 else 's'} published."

            return self.render(
                request,
                redirect_url=self.get_done_url(),
                message=message,
                collect_render_data=False,
            )
        return self.render(
            request, queryset=queryset, publish_form=form, action="Publish"
        )


class UnPublishActionView(PublishActionView):
    """
    View for unpublishing one or more objects. \
    Inherits from PublishActionView.

    :param default_template: Defaults to cms/publish_action.html.
    :param redirect_to_view: Defaults to 'Unpublish selected items'
    """

    short_description = "Unpublish selected items"
    action_name = "Unpublish"

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests. Passes the
        following arguments to the context:

        * **queryset** - The queryset of objects to unpublish.
        """

        queryset = self.get_selected(request)
        return self.render(request, queryset=queryset, action="Unpublish")

    def process_action(self, request, queryset):
        """
        Unpublishes the selected objects by calling the object's \
        unpublish method. The action is logged and the user is \
        notified with a message.

        Returns a 'render redirect' to the result of the \
        `get_done_url` method.
        """
        count = 0
        for obj in queryset:
            count += 1
            obj.unpublish()
            object_url = self.get_object_url(obj)
            self.log_action(obj, CMSLog.UNPUBLISH, url=object_url)
        url = self.get_done_url()
        msg = f"{count} object{'' if count == 1 else 's'} unpublished."
        return self.render(
            request,
            redirect_url=url,
            message=msg,
            collect_render_data=False,
        )


# deprecated views
class PublishView(ModelCMSMixin, SingleObjectMixin, ModelCMSView):
    """
    View for publishing an object. Inherits from
    ModelCMSMixin, SingleObjectMixin, ModelCMSView.
    Assumes the given model is versionable.

    :param default_template: Defaults to 'cms/publish.html'.
    :param redirect_to_view: Defaults to 'edit'.
    """

    default_template = "cms/publish.html"
    redirect_to_view = "edit"
    form = WhenForm

    def get_object(self):
        """
        Get the object for publishing
        Raises a http404 error if the object is not found.
        """
        obj = super().get_object()

        if not obj or not hasattr(obj, "publish"):
            raise http.Http404

        return obj

    def get_object_url(self):
        """
        Returns the url to link to the object
        The get_view_url will be called on the current bundle using
        'edit` as the view name.
        """
        return self.bundle.get_view_url("edit", self.request.user, {}, self.kwargs)

    def get_done_url(self):
        """
        Returns the url to redirect to after a successful update.
        The get_view_url will be called on the current bundle using
        `self.redirect_to_view` as the view name.
        """
        url = self.bundle.get_view_url(
            self.redirect_to_view, self.request.user, {}, self.kwargs
        )
        return self.customized_return_url(url)

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests. Passes the
        following arguments to the context:

        * **obj** - The object to publish.
        * **form** - An instance of WhenForm.
        * **done_url** - The result of the `get_done_url` method.
        """

        self.object = self.get_object()
        return self.render(
            request, obj=self.object, form=self.form(), done_url=self.get_done_url()
        )

    def post(self, request, *args, **kwargs):
        """
        Method for handling POST requests. Publishes
        the object passing the value of 'when' to the object's
        publish method. The object's `purge_archives` method
        is also called to limit the number of old items
        that we keep around. The action is logged as either
        'published' or 'scheduled' depending on the value of
        'when', and the user is notified with a message.

        Returns a 'render redirect' to the result of the
        `get_done_url` method.
        """

        self.object = self.get_object()
        form = self.form()
        url = self.get_done_url()
        if request.POST.get("publish"):
            form = self.form(request.POST)
            if form.is_valid():
                when = form.cleaned_data.get("when")
                self.object.publish(user=request.user, when=when)
                self.object.purge_archives()
                object_url = self.get_object_url()
                if self.object.state == self.object.PUBLISHED:
                    self.log_action(self.object, CMSLog.PUBLISH, url=object_url)
                else:
                    self.log_action(self.object, CMSLog.SCHEDULE, url=object_url)

                message = f"{self.object} {self.object.state}"

                return self.render(
                    request,
                    redirect_url=url,
                    message=message,
                    obj=self.object,
                    collect_render_data=False,
                )
        return self.render(request, obj=self.object, form=form, done_url=url)


class UnPublishView(PublishView):
    """
    View for unpublishing an object. Inherits from PublishView.

    :param default_template: 'cms/unpublish.html'.
    :param redirect_to_view: 'edit'.
    """

    default_template = "cms/unpublish.html"
    redirect_to_view = "edit"

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests. Passes the
        following arguments to the context:

        * **obj** - The object to publish
        * **done_url** - The result of the `get_done_url` method
        """

        self.object = self.get_object()
        return self.render(request, obj=self.object, done_url=self.get_done_url())

    def post(self, request, *args, **kwargs):
        """
        Method for handling POST requests. Unpublishes the
        the object by calling the object's unpublish method.
        The action is logged, the user is notified with a
        message. Returns a 'render redirect' to the result of
        the `get_done_url` method.
        """

        self.object = self.get_object()
        url = self.get_done_url()
        if request.POST.get("unpublish"):
            self.object.unpublish()
            object_url = self.get_object_url()
            self.log_action(self.object, CMSLog.UNPUBLISH, url=object_url)
            msg = f"{self.object} unpublished"
            return self.render(
                request,
                redirect_url=url,
                message=msg,
                obj=self.object,
                collect_render_data=False,
            )

        return self.render(request, obj=self.object, done_url=url)


class DeleteView(ModelCMSMixin, SingleObjectMixin, ModelCMSView):
    """
    View for deleting an object.
    Inherits from ModelCMSMixin, SingleObjectMixin and ModelCMSView.
    Used through out the CMS as the default object_header view so
    this view by default adds an `object_header` renderer that
    uses the `object_header_tmpl` template.

    :param default_template: Defaults to 'cms/delete.html'.
    :param redirect_to_view: Defaults to 'main_list'.
    """

    default_template = "cms/delete.html"
    redirect_to_view = "main_list"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.renders["object_header"] = renders.RenderString(
            template=self.object_header_tmpl
        )

    def get_done_url(self):
        """
        Returns the url to redirect to after a successful update.
        The get_view_url will be called on the current bundle using
        `self.redirect_to_view` as the view name.
        """
        data = dict(self.kwargs)
        data.pop(self.slug_url_kwarg, None)
        url = self.bundle.get_view_url(self.redirect_to_view, self.request.user, data)
        return self.customized_return_url(url)

    def get_object(self):
        """
        Get the object for previewing.
        Raises a http404 error if the object is not found.
        """
        obj = super().get_object()

        if not obj:
            raise http.Http404

        return obj

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests.
        Calls the `render` method with the following
        items in context:

        * **obj** - The obj being deleted.
        """

        self.object = self.get_object()
        return self.render(request, obj=self.object)

    def post(self, request, *args, **kwargs):
        """
        Method for handling POST requests.
        Deletes the object. Successful deletes are logged.
        Returns a 'render redirect' to the result of the
        `get_done_url` method.

        If a ProtectedError is raised, the `render` method
        is called with message explaining the error added
        to the context as `protected`.
        """

        self.object = self.get_object()
        msg = None
        if request.POST.get("delete"):
            try:
                with transaction.commit_on_success():
                    self.log_action(self.object, CMSLog.DELETE)
                    msg = f"{self.object} deleted"
                    self.object.delete()
            except ProtectedError as e:
                protected = []
                for x in e.protected_objects:
                    if hasattr(x, "delete_blocked_message"):
                        protected.append(x.delete_blocked_message())
                    else:
                        protected.append(f"{x._meta.verbose_name}: {x}")
                return self.render(request, obj=self.object, protected=protected)

        return self.render(
            request,
            redirect_url=self.get_done_url(),
            obj=self.object,
            message=msg,
            collect_render_data=False,
        )



from .. import Provider as AddressProvider


class Provider(AddressProvider):
    city_suffixes = ("市", "县")
    city_formats = ("{{city_name}}{{city_suffix}}",
                    "{{first_name}}{{city_suffix}}")

    district_formats = ("{{district}}区",)

    building_number_formats = ("?座",)
    postcode_formats = ("%#####",)

    street_suffixes = ("街", "路")
    street_name_formats = ("{{city_name}}{{street_suffix}}",
                           "{{last_name}}{{street_suffix}}")
    street_address_formats = ("{{street_name}}{{building_number}}",)

    address_formats = (
        "{{province}}{{city}}{{district}}{{street_address}} {{postcode}}",)

    provinces = (
        "北京市", "上海市", "天津市", "重庆市",
        "内蒙古自治区", "山西省", "河北省", "吉林省", "江苏省", "辽宁省", "黑龙江省",
        "安徽省", "山东省", "浙江省", "江西省", "福建省", "湖南省", "湖北省",
        "河南省", "广东省", "广西壮族自治区", "贵州省", "海南省", "四川省", "云南省",
        "陕西省", "甘肃省", "宁夏回族自治区", "青海省", "新疆维吾尔自治区", "西藏自治区",
        "台湾省", "香港特别行政区", "澳门特别行政区",
    )
    districts = (
        "西夏", "永川", "秀英", "高港", "清城", "兴山", "锡山", "清河",
        "龙潭", "华龙", "海陵", "滨城", "东丽", "高坪", "沙湾", "平山",
        "城北", "海港", "沙市", "双滦", "长寿", "山亭", "南湖", "浔阳",
        "南长", "友好", "安次", "翔安", "沈河", "魏都", "西峰", "萧山",
        "金平", "沈北新", "孝南", "上街", "城东", "牧野", "大东",
        "白云", "花溪", "吉区", "新城", "怀柔", "六枝特", "涪城",
        "清浦", "南溪", "淄川", "高明", "东城", "崇文", "朝阳", "大兴",
        "房山", "门头沟", "黄浦", "徐汇", "静安", "普陀", "闵行", "和平",
        "蓟州", "永川", "长寿", "璧山", "合川", "梁平", "丰都", "江北",
    )
    cities = (
        "北京", "上海", "天津", "重庆", "哈尔滨", "长春", "沈阳", "呼和浩特",
        "石家庄", "乌鲁木齐", "兰州", "西宁", "西安", "银川", "郑州", "济南", "太原",
        "合肥", "武汉", "长沙", "南京", "成都", "贵阳", "昆明", "南宁", "拉萨",
        "杭州", "南昌", "广州", "福州", "台北", "海口", "香港", "澳门", "通辽",
        "兴安盟", "太原", "辛集", "邯郸", "沈阳", "辽阳", "兴城", "北镇", "阜新",
        "哈尔滨", "齐齐哈尔", "淮安", "张家港", "海门", "六安", "巢湖", "马鞍山",
        "永安", "宁德", "嘉禾", "荆门", "潜江", "大冶", "宜都", "佛山", "深圳",
        "潮州", "惠州", "汕尾", "东莞", "梧州", "柳州", "合山", "六盘水", "关岭")
    countries = (
        "阿富汗", "阿拉斯加", "阿尔巴尼亚", "阿尔及利亚", "安道尔", "安哥拉", "安圭拉岛英", "安提瓜和巴布达",
        "阿根廷", "亚美尼亚", "阿鲁巴岛", "阿森松", "澳大利亚", "奥地利", "阿塞拜疆", "巴林", "孟加拉国",
        "巴巴多斯", "白俄罗斯", "比利时", "伯利兹", "贝宁", "百慕大群岛", "不丹", "玻利维亚", "波斯尼亚和黑塞哥维那",
        "博茨瓦纳", "巴西", "保加利亚", "布基纳法索", "布隆迪", "喀麦隆", "加拿大", "加那利群岛", "佛得角",
        "开曼群岛", "中非", "乍得", "智利", "圣诞岛", "科科斯岛", "哥伦比亚", "巴哈马国", "多米尼克国", "科摩罗",
        "刚果", "科克群岛", "哥斯达黎加", "克罗地亚", "古巴", "塞浦路斯", "捷克", "丹麦", "迪戈加西亚岛", "吉布提",
        "多米尼加共和国", "厄瓜多尔", "埃及", "萨尔瓦多", "赤道几内亚", "厄立特里亚", "爱沙尼亚", "埃塞俄比亚", "福克兰群岛",
        "法罗群岛", "斐济", "芬兰", "法国", "法属圭亚那", "法属波里尼西亚", "加蓬", "冈比亚", "格鲁吉亚", "德国", "加纳",
        "直布罗陀", "希腊", "格陵兰岛", "格林纳达", "瓜德罗普岛", "关岛", "危地马拉", "几内亚", "几内亚比绍", "圭亚那",
        "海地", "夏威夷", "洪都拉斯", "匈牙利", "冰岛", "印度", "印度尼西亚", "伊郎", "伊拉克", "爱尔兰", "以色列",
        "意大利", "科特迪瓦", "牙买加", "日本", "约旦", "柬埔塞", "哈萨克斯坦", "肯尼亚", "基里巴斯", "朝鲜", "韩国",
        "科威特", "吉尔吉斯斯坦", "老挝", "拉脱维亚", "黎巴嫩", "莱索托", "利比里亚", "利比亚", "列支敦士登", "立陶宛",
        "卢森堡", "马其顿", "马达加斯加", "马拉维", "马来西亚", "马尔代夫", "马里", "马耳他", "马里亚纳群岛", "马绍尔群岛",
        "马提尼克", "毛里塔尼亚", "毛里求斯", "马约特岛", "墨西哥", "密克罗尼西亚", "中途岛", "摩尔多瓦", "摩纳哥", "蒙古",
        "蒙特塞拉特岛", "摩洛哥", "莫桑比克", "缅甸", "纳米比亚", "瑙鲁", "尼泊尔", "荷兰", "荷属安的列斯群岛", "新喀里多尼亚群岛",
        "新西兰", "尼加拉瓜", "尼日尔", "尼日利亚", "纽埃岛", "诺福克岛", "挪威", "阿曼", "帕劳", "巴拿马", "巴布亚新几内亚",
        "巴拉圭", "秘鲁", "菲律宾", "波兰", "葡萄牙", "巴基斯坦", "波多黎各", "卡塔尔", "留尼汪岛", "罗马尼亚", "俄罗斯",
        "卢旺达", "东萨摩亚", "西萨摩亚", "圣马力诺", "圣皮埃尔岛及密克隆岛", "圣多美和普林西比", "沙特阿拉伯", "塞内加尔",
        "塞舌尔", "新加坡", "斯洛伐克", "斯洛文尼亚", "所罗门群岛", "索马里", "南非", "西班牙", "斯里兰卡", "圣克里斯托弗和尼维斯",
        "圣赫勒拿", "圣卢西亚", "圣文森特岛", "苏丹", "苏里南", "斯威士兰", "瑞典", "瑞士", "叙利亚", "塔吉克斯坦", "坦桑尼亚",
        "泰国", "阿拉伯联合酋长国", "多哥", "托克劳群岛", "汤加", "特立尼达和多巴哥", "突尼斯", "土耳其", "土库曼斯坦",
        "特克斯和凯科斯群岛", "图瓦卢", "美国", "乌干达", "乌克兰", "英国", "乌拉圭", "乌兹别克斯坦", "瓦努阿图",
        "梵蒂冈", "委内瑞拉", "越南", "维尔京群岛", "维尔京群岛和圣罗克伊", "威克岛", "瓦里斯和富士那群岛", "西撒哈拉",
        "也门", "南斯拉夫", "扎伊尔", "赞比亚", "桑给巴尔", "津巴布韦", "中华人民共和国", "中国",
    )

    def building_number(self):
        return self.lexify(self.random_element(self.building_number_formats))

    def city_name(self):
        return self.random_element(self.cities)

    def province(self):
        return self.random_element(self.provinces)

    def district(self):
        return self.random_element(self.districts)



from .. import Provider as DateTimeProvider


class Provider(DateTimeProvider):

    def day_of_week(self):
        day = self.date('%w')
        DAY_NAMES = {
            "0": "일요일",
            "1": "월요일",
            "2": "화요일",
            "3": "수요일",
            "4": "목요일",
            "5": "금요일",
            "6": "토요일",
        }
        return DAY_NAMES[day]

    def month_name(self):
        month = self.month()
        MONTH_NAMES = {
            "01": "1월",
            "02": "2월",
            "03": "3월",
            "04": "4월",
            "05": "5월",
            "06": "6월",
            "07": "7월",
            "08": "8월",
            "09": "9월",
            "10": "10월",
            "11": "11월",
            "12": "12월",
        }
        return MONTH_NAMES[month]



from django import VERSION as DJANGO_VERSION
from django.contrib import auth
from django.contrib.auth import models as auth_app
from django.contrib.auth.models import Group
from django.db.models.signals import post_migrate

from .. import settings


def ensure_groups_sync(app, created_models, verbosity, **kwargs):
    for group in settings.BASE_GROUPS:
        Group.objects.get_or_create(name=group)


def ensure_groups(sender=None, **kwargs):
    for group in settings.BASE_GROUPS:
        Group.objects.get_or_create(name=group)


post_migrate.connect(ensure_groups, sender=auth_app, dispatch_uid="ensure_groups")



import fnmatch
import os

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def matches_patterns(path, patterns=None):
    """
    Return True or False depending on whether the ``path`` should be
    ignored (if it matches any pattern in ``ignore_patterns``).
    """
    return any(fnmatch.fnmatchcase(path, pattern) for pattern in (patterns or []))


def get_files(storage, ignore_patterns=None, location=''):
    """
    Recursively walk the storage directories yielding the paths
    of all files that should be copied.
    """
    if ignore_patterns is None:
        ignore_patterns = []
    directories, files = storage.listdir(location)
    for fn in files:
        # Match only the basename.
        if matches_patterns(fn, ignore_patterns):
            continue
        if location:
            fn = os.path.join(location, fn)
            # Match the full file path.
            if matches_patterns(fn, ignore_patterns):
                continue
        yield fn
    for dir in directories:
        if matches_patterns(dir, ignore_patterns):
            continue
        if location:
            dir = os.path.join(location, dir)
        yield from get_files(storage, ignore_patterns, dir)


def check_settings(base_url=None):
    """
    Check if the staticfiles settings have sane values.
    """
    if base_url is None:
        base_url = settings.STATIC_URL
    if not base_url:
        raise ImproperlyConfigured(
            "You're using the staticfiles app "
            "without having set the required STATIC_URL setting.")
    if settings.MEDIA_URL == base_url:
        raise ImproperlyConfigured("The MEDIA_URL and STATIC_URL "
                                   "settings must have different values")
    if (settings.DEBUG and settings.MEDIA_URL and settings.STATIC_URL and
            settings.MEDIA_URL.startswith(settings.STATIC_URL)):
        raise ImproperlyConfigured(
            "runserver can't serve media if MEDIA_URL is within STATIC_URL."
        )
    if ((settings.MEDIA_ROOT and settings.STATIC_ROOT) and
            (settings.MEDIA_ROOT == settings.STATIC_ROOT)):
        raise ImproperlyConfigured("The MEDIA_ROOT and STATIC_ROOT "
                                   "settings must have different values")



from django.utils.safestring import mark_safe

try:
    from ..cms import bundles, cms_bundles, views
    from ..cms.sites import site
except ValueError:
    from cms import site, bundles, views, cms_bundles

from . import forms, get_asset_model, models
from .views import AssetFormView, AssetListView, CropVersionView, CropView


def preview(obj):
    if obj.type in [obj.IMAGE, obj.SVG_IMAGE]:
        thumbnail = obj.file.admin_url()
        if thumbnail:
            return mark_safe('<img src="{0}" />'.format(thumbnail))

    return ""


# class CropBundle(bundles.BlankBundle):
#     navigation = bundles.PARENT
#     object_view = bundles.PARENT

#     edit_asset = bundles.URLAlias(bundle_attr=bundles.PARENT, alias_to="edit")

#     main = views.ListView(
#         allow_empty=False,
#         base_template="cms/partial.html",
#         base_filter_kwargs={"editable": True},
#     )
#     edit = CropVersionView()

#     class Meta:
#         parent_field = "image"
#         model = models.ImageDetail
#         edit_regex_base = "(?P<%(group_name)s_pk>[a-zA-z0-9_.-]+)/%(attname)s/"


class AssetBundle(bundles.Bundle):

    main = AssetListView(
        display_fields=(preview, "title", "user_filename", "modified", "type")
    )
    add = AssetFormView(force_add=True, form_class=forms.UploadAssetForm)
    edit = AssetFormView(form_class=forms.UpdateAssetCropForm)#form_class=forms.UpdateAssetForm)
    # crop = CropView(default_template="assets/crop.html")
    # crops = CropBundle.as_subbundle(name="crops")

    class Meta:
        #item_views = ("edit", "crop", "crops", "delete")
        item_views = ("edit", "delete")
        primary_model_bundle = True
        model = get_asset_model()


class EmbedView(cms_bundles.EmbedView):
    def get(self, request, *args, **kwargs):
        tags = request.GET.get("tags")
        bundle = self.bundle.admin_site.get_bundle_for_model(get_asset_model())
        api_link = ""
        if bundle:
            api_link = bundle.get_view_url("main", request.user)
            if api_link:
                api_link = "{0}?type=choices".format(api_link)

        return self.render(request, tags=tags, api_link=api_link)


class WYSIWYG(cms_bundles.WYSIWYG):
    main = EmbedView(default_template="cms/insert_media.html")


site.unregister("wysiwyg")
site.register("wysiwyg", WYSIWYG(name="wysiwyg"), 21)
site.register("assets", AssetBundle(name="assets"), 20)



from django.shortcuts import redirect

from .middleware import SESSION_KEY


def switch_state(request):
    """
    Switch the default version state in
    the session.
    """

    if request.session.get(SESSION_KEY):
        request.session[SESSION_KEY] = False
    else:
        request.session[SESSION_KEY] = True

    # Get redirect location
    # Don't go to non local paths
    url = request.GET.get("redirect_to", "/")
    if url.startswith("http"):
        url = "/"
    return redirect(url)



# -*- coding: utf-8 -*-

# The following comment should be removed at some point in the future.
# mypy: disallow-untyped-defs=False

from __future__ import absolute_import

import logging
import os
import shutil

from pip._internal.cache import WheelCache
from pip._internal.cli import cmdoptions
from pip._internal.cli.req_command import RequirementCommand
from pip._internal.exceptions import CommandError, PreviousBuildDirError
from pip._internal.req import RequirementSet
from pip._internal.req.req_tracker import get_requirement_tracker
from pip._internal.utils.misc import ensure_dir, normalize_path
from pip._internal.utils.temp_dir import TempDirectory
from pip._internal.utils.typing import MYPY_CHECK_RUNNING
from pip._internal.wheel_builder import build, should_build_for_wheel_command

if MYPY_CHECK_RUNNING:
    from optparse import Values
    from typing import Any, List


logger = logging.getLogger(__name__)


class WheelCommand(RequirementCommand):
    """
    Build Wheel archives for your requirements and dependencies.

    Wheel is a built-package format, and offers the advantage of not
    recompiling your software during every install. For more details, see the
    wheel docs: https://wheel.readthedocs.io/en/latest/

    Requirements: setuptools>=0.8, and wheel.

    'pip wheel' uses the bdist_wheel setuptools extension from the wheel
    package to build individual wheels.

    """

    usage = """
      %prog [options] <requirement specifier> ...
      %prog [options] -r <requirements file> ...
      %prog [options] [-e] <vcs project url> ...
      %prog [options] [-e] <local project path> ...
      %prog [options] <archive url/path> ..."""

    def __init__(self, *args, **kw):
        super(WheelCommand, self).__init__(*args, **kw)

        cmd_opts = self.cmd_opts

        cmd_opts.add_option(
            '-w', '--wheel-dir',
            dest='wheel_dir',
            metavar='dir',
            default=os.curdir,
            help=("Build wheels into <dir>, where the default is the "
                  "current working directory."),
        )
        cmd_opts.add_option(cmdoptions.no_binary())
        cmd_opts.add_option(cmdoptions.only_binary())
        cmd_opts.add_option(cmdoptions.prefer_binary())
        cmd_opts.add_option(
            '--build-option',
            dest='build_options',
            metavar='options',
            action='append',
            help="Extra arguments to be supplied to 'setup.py bdist_wheel'.",
        )
        cmd_opts.add_option(cmdoptions.no_build_isolation())
        cmd_opts.add_option(cmdoptions.use_pep517())
        cmd_opts.add_option(cmdoptions.no_use_pep517())
        cmd_opts.add_option(cmdoptions.constraints())
        cmd_opts.add_option(cmdoptions.editable())
        cmd_opts.add_option(cmdoptions.requirements())
        cmd_opts.add_option(cmdoptions.src())
        cmd_opts.add_option(cmdoptions.ignore_requires_python())
        cmd_opts.add_option(cmdoptions.no_deps())
        cmd_opts.add_option(cmdoptions.build_dir())
        cmd_opts.add_option(cmdoptions.progress_bar())

        cmd_opts.add_option(
            '--global-option',
            dest='global_options',
            action='append',
            metavar='options',
            help="Extra global options to be supplied to the setup.py "
            "call before the 'bdist_wheel' command.")

        cmd_opts.add_option(
            '--pre',
            action='store_true',
            default=False,
            help=("Include pre-release and development versions. By default, "
                  "pip only finds stable versions."),
        )

        cmd_opts.add_option(cmdoptions.no_clean())
        cmd_opts.add_option(cmdoptions.require_hashes())

        index_opts = cmdoptions.make_option_group(
            cmdoptions.index_group,
            self.parser,
        )

        self.parser.insert_option_group(0, index_opts)
        self.parser.insert_option_group(0, cmd_opts)

    def run(self, options, args):
        # type: (Values, List[Any]) -> None
        cmdoptions.check_install_build_global(options)

        session = self.get_default_session(options)

        finder = self._build_package_finder(options, session)
        build_delete = (not (options.no_clean or options.build_dir))
        wheel_cache = WheelCache(options.cache_dir, options.format_control)

        options.wheel_dir = normalize_path(options.wheel_dir)
        ensure_dir(options.wheel_dir)

        with get_requirement_tracker() as req_tracker, TempDirectory(
            options.build_dir, delete=build_delete, kind="wheel"
        ) as directory:

            requirement_set = RequirementSet()

            try:
                self.populate_requirement_set(
                    requirement_set, args, options, finder, session,
                    wheel_cache
                )

                preparer = self.make_requirement_preparer(
                    temp_build_dir=directory,
                    options=options,
                    req_tracker=req_tracker,
                    session=session,
                    finder=finder,
                    wheel_download_dir=options.wheel_dir,
                    use_user_site=False,
                )

                resolver = self.make_resolver(
                    preparer=preparer,
                    finder=finder,
                    options=options,
                    wheel_cache=wheel_cache,
                    ignore_requires_python=options.ignore_requires_python,
                    use_pep517=options.use_pep517,
                )

                self.trace_basic_info(finder)

                resolver.resolve(requirement_set)

                reqs_to_build = [
                    r for r in requirement_set.requirements.values()
                    if should_build_for_wheel_command(r)
                ]

                # build wheels
                build_successes, build_failures = build(
                    reqs_to_build,
                    wheel_cache=wheel_cache,
                    build_options=options.build_options or [],
                    global_options=options.global_options or [],
                )
                for req in build_successes:
                    assert req.link and req.link.is_wheel
                    assert req.local_file_path
                    # copy from cache to target directory
                    try:
                        shutil.copy(req.local_file_path, options.wheel_dir)
                    except OSError as e:
                        logger.warning(
                            "Building wheel for %s failed: %s",
                            req.name, e,
                        )
                        build_failures.append(req)
                if len(build_failures) != 0:
                    raise CommandError(
                        "Failed to build one or more wheels"
                    )
            except PreviousBuildDirError:
                options.no_clean = True
                raise
            finally:
                if not options.no_clean:
                    requirement_set.cleanup_files()
                    wheel_cache.cleanup()



from django.db.models.sql import compiler


class SQLCompiler(compiler.SQLCompiler):
    def as_subquery_condition(self, alias, columns, compiler):
        qn = compiler.quote_name_unless_alias
        qn2 = self.connection.ops.quote_name
        sql, params = self.as_sql()
        return '(%s) IN (%s)' % (', '.join('%s.%s' % (qn(alias), qn2(column)) for column in columns), sql), params


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):
    pass


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    pass


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    pass


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    pass



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from . import settings as accounts_settings

attrs_dict = {"class": "required"}

USERNAME_RE = r"^[\.\w]+$"


class PasswordLinkWidget(forms.Widget):
    def render(self, name, value, attrs=None, renderer=None):
        return mark_safe('<a href="../password/" id="password-link">Change</a>')


class UserForm(forms.ModelForm):
    password = forms.CharField(required=False, widget=PasswordLinkWidget())
    groups = forms.ModelMultipleChoiceField(
        required=False, queryset=Group.objects.all().order_by("name")
    )

    def clean(self):
        if self.is_valid():
            self.cleaned_data["password"] = self.instance.password
            return self.cleaned_data

        if hasattr(self, "cleaned_data"):
            return self.cleaned_data
        else:
            return {}

    class Meta:
        model = get_user_model()
        exclude = []


class SignupModelForm(forms.ModelForm):
    """
    Form for creating a new user account.
    Validates that the requested username and e-mail is not already in use.
    Also requires the password to be entered twice.
    """

    first_name = forms.CharField(label=_("First name"), max_length=30)
    last_name = forms.CharField(label=_("Last name"), max_length=30)
    username = forms.RegexField(
        regex=USERNAME_RE,
        max_length=30,
        widget=forms.TextInput(attrs=attrs_dict),
        label=_("Username"),
        error_messages={
            "invalid": _(
                "Username must contain only letters, numbers, dots and underscores."
            )
        },
    )
    email = forms.EmailField(
        widget=forms.TextInput(attrs=dict(attrs_dict, maxlength=75)), label=_("Email")
    )
    password1 = forms.CharField(
        widget=forms.PasswordInput(attrs=attrs_dict, render_value=False),
        label=_("Create password"),
    )
    password2 = forms.CharField(
        widget=forms.PasswordInput(attrs=attrs_dict, render_value=False),
        label=_("Repeat password"),
    )

    def clean_username(self):
        """
        Validate that the username is alphanumeric and is not already in use.
        Also validates that the username is not listed in
        ACCOUNTS_FORBIDDEN_USERNAMES list.
        """
        try:
            get_user_model().objects.get(username__iexact=self.cleaned_data["username"])
        except get_user_model().DoesNotExist:
            pass
        else:
            raise forms.ValidationError(_("This username is already taken."))
        if (
            self.cleaned_data["username"].lower()
            in accounts_settings.ACCOUNTS_FORBIDDEN_USERNAMES
        ):
            raise forms.ValidationError(_("This username is not allowed."))
        return self.cleaned_data["username"]

    def clean_email(self):
        """
        Validate that the e-mail address is unique.
        """
        if get_user_model().objects.filter(email__iexact=self.cleaned_data["email"]):
            raise forms.ValidationError(
                _("This email is already in use. Please supply a different email.")
            )
        return self.cleaned_data["email"]

    def clean(self):
        """
        Validates that the values entered into the two password fields match.
        Note that an error here will end up in ``non_field_errors()`` because
        it doesn't apply to a single field.
        """
        if "password1" in self.cleaned_data and "password2" in self.cleaned_data:
            if self.cleaned_data["password1"] != self.cleaned_data["password2"]:
                raise forms.ValidationError(_("The two password fields didn't match."))
        return self.cleaned_data

    def save(self):
        """ Creates a new user and account. Returns the newly created user. """
        username, email, password, first_name, last_name = (
            self.cleaned_data["username"],
            self.cleaned_data["email"],
            self.cleaned_data["password1"],
            self.cleaned_data["first_name"],
            self.cleaned_data["last_name"],
        )

        new_user = get_user_model()(
            username=username, email=email, first_name=first_name, last_name=last_name
        )
        new_user.set_password(password)
        new_user.save()
        return new_user

    class Meta:
        model = get_user_model()
        fields = (
            "first_name",
            "last_name",
            "username",
            "email",
            "password1",
            "password2",
        )



import json

from django.conf import settings
from django.contrib.messages.storage.base import BaseStorage
from django.contrib.messages.storage.cookie import (
    MessageDecoder, MessageEncoder,
)


class SessionStorage(BaseStorage):
    """
    Store messages in the session (that is, django.contrib.sessions).
    """
    session_key = '_messages'

    def __init__(self, request, *args, **kwargs):
        assert hasattr(request, 'session'), "The session-based temporary "\
            "message storage requires session middleware to be installed, "\
            "and come before the message middleware in the "\
            "MIDDLEWARE%s list." % ("_CLASSES" if settings.MIDDLEWARE is None else "")
        super().__init__(request, *args, **kwargs)

    def _get(self, *args, **kwargs):
        """
        Retrieve a list of messages from the request's session. This storage
        always stores everything it is given, so return True for the
        all_retrieved flag.
        """
        return self.deserialize_messages(self.request.session.get(self.session_key)), True

    def _store(self, messages, response, *args, **kwargs):
        """
        Store a list of messages to the request's session.
        """
        if messages:
            self.request.session[self.session_key] = self.serialize_messages(messages)
        else:
            self.request.session.pop(self.session_key, None)
        return []

    def serialize_messages(self, messages):
        encoder = MessageEncoder(separators=(',', ':'))
        return encoder.encode(messages)

    def deserialize_messages(self, data):
        if data and isinstance(data, str):
            return json.loads(data, cls=MessageDecoder)
        return data



from django.db import models
from django.db.models.query import QuerySet
from scarlet.versioning.models import Cloneable, VersionView
from scarlet.versioning.fields import FKToVersion
from scarlet.pagebuilder import settings, base_models

"""
This file can be copied into a project's app folder as models.py for
default implementation of Pagebuilder.
"""


class Page(VersionView, base_models.BasePage):
    _clone_related = [
        "hero_modules",
        "two_column_modules",
        "icon_list_modules",
        "header_modules",
        "faq_modules",
        "location_modules",
        "gallery_modules",
        "carousel_modules",
    ]
    versioned_unique = ("slug",)

    @property
    def site_name(self):
        return settings.SITE_NAME


class HeroModule(Cloneable, base_models.BaseHeroModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="hero_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    styles = (("left", "left"), ("right", "right"), ("center", "center"))
    behaviors = (("horizontal", "horizontal"), ("vertical", "vertical"))


class TwoColumnModule(Cloneable, base_models.BaseTwoColumnModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="two_column_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )


class IconListManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return super().get_query_set().prefetch_related("icon_list_items")


class IconListModule(Cloneable, base_models.BaseIconListModule):
    _clone_related = ("icon_list_items",)
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="icon_list_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    objects = IconListManager()

    @property
    def items(self) -> QuerySet:
        return self.icon_list_items.all()


class IconListItem(Cloneable, base_models.BaseIconListItem):
    module = models.ForeignKey(
        settings.ICON_LIST_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="icon_list_items",
    )


class HeaderModule(Cloneable, base_models.BaseHeaderModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="header_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    styles = (("left", "left"), ("right", "right"), ("center", "center"))
    behaviors = (("horizontal", "horizontal"), ("vertical", "vertical"))


class FAQManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return super().get_query_set().prefetch_related("faq_items")


class FAQModule(Cloneable, base_models.BaseFAQModule):
    _clone_related = ("faq_items",)
    objects = FAQManager()

    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="faq_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )

    @property
    def items(self) -> QuerySet:
        return self.faq_items.all()


class FAQItem(Cloneable, base_models.BaseFAQItem):
    module = models.ForeignKey(
        settings.FAQ_MODULE_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="faq_items",
    )


class LocationManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return super().get_query_set().prefetch_related("location_items")


class LocationModule(Cloneable, base_models.BaseLocationModule):
    _clone_related = ("location_items",)
    objects = LocationManager()

    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="location_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )

    @property
    def items(self) -> QuerySet:
        return self.location_items.all()


class LocationItem(Cloneable, base_models.BaseLocationItem):
    module = models.ForeignKey(
        settings.LOCATION_MODULE_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="location_items",
    )


class GalleryManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return super().get_query_set().prefetch_related("gallery_images")


class ImageGalleryModule(Cloneable, base_models.BaseImageGalleryModule):
    _clone_related = ("gallery_images",)
    objects = GalleryManager()

    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="gallery_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )

    @property
    def items(self) -> QuerySet:
        return self.gallery_images.all()


class GalleryImage(Cloneable, base_models.BaseGalleryImage):
    module = models.ForeignKey(
        settings.GALLERY_MODULE_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="gallery_images",
    )


class CarouselManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return super().get_query_set().prefetch_related("carousel_items")


class CarouselModule(Cloneable, base_models.BaseCarouselModule):
    _clone_related = ("carousel_items",)
    objects = CarouselManager()

    @property
    def items(self) -> QuerySet:
        return self.carousel_items.all()

    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="carousel_modules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )


class CarouselItem(Cloneable, base_models.BaseCarouselItem):
    module = models.ForeignKey(
        settings.CAROUSEL_MODULE_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="carousel_items",
    )



"""
Wrapper for loading templates from a plain Python dict.
"""

from django.template import Origin, TemplateDoesNotExist

from .base import Loader as BaseLoader


class Loader(BaseLoader):

    def __init__(self, engine, templates_dict):
        self.templates_dict = templates_dict
        super().__init__(engine)

    def get_contents(self, origin):
        try:
            return self.templates_dict[origin.name]
        except KeyError:
            raise TemplateDoesNotExist(origin)

    def get_template_sources(self, template_name):
        yield Origin(
            name=template_name,
            template_name=template_name,
            loader=self,
        )



import os
import random
from functools import wraps

from django.core.cache import cache
from django.core.validators import get_available_image_extensions

from . import settings


IMAGE_EXTENSIONS = get_available_image_extensions()


def partial(func, *parameters, **kparms):
    @wraps(func)
    def wrapped(*args, **kw):
        kw.update(kparms)
        return func(*(args + parameters), **kw)

    return wrapped


def assets_dir(instance, filename):
    name, ext = os.path.splitext(filename)
    if instance.rename_file():
        name = instance.slug
    if settings.DIRECTORY:
        return "/".join([settings.DIRECTORY, name + ext])
    else:
        return instance.slug + ext


def get_size_filename(filename, size_name):
    filename, ext = os.path.splitext(filename)
    return filename + "_" + size_name + ext


def get_cache_bust_version(url):
    key = "cbversion.{0}".format(url)
    value = cache.get(key)
    if not value:
        # We could look it up but for now just make it up
        value = update_cache_bust_version(url, random.randint(0, 60))
    return value


def update_cache_bust_version(url, value=None):
    key = "cbversion.{0}".format(url)
    if not value:
        value = cache.get(key)
    if value:
        value = int(value) + 1
    else:
        value = 1
    cache.set(key, value, 60 * 60 * 24 * 60)
    return value



from django.core.checks import Error


BAD_CONFIG_ERROR = Error(
    'Error while parsing WEBPACK_LOADER configuration',
    hint='Is WEBPACK_LOADER config valid?',
    obj='django.conf.settings.WEBPACK_LOADER',
    id='django-webpack-loader.E001',
)



from django.urls import path

from .views import (
    DummyCacheView,
    DummyCacheListView,
    GetCacheVersionNotImplementedView,
    GetCacheVersionNotImplementedListView,
)

urlpatterns = [
    path("dummy_cache_view/", DummyCacheView.as_view(),),
    path(
        "get_cache_version_not_implemented_view/",
        GetCacheVersionNotImplementedView.as_view(),
    ),
    path("dummy_cache_list_view/", DummyCacheListView.as_view(),),
    path(
        "get_cache_version_not_implemented_list_view/",
        GetCacheVersionNotImplementedListView.as_view(),
    ),
]



from .. import Provider as PersonProvider


class Provider(PersonProvider):
    formats_female = (
        '{{last_name}} {{first_name_female}}',
    )

    formats_male = (
        '{{last_name}} {{first_name_male}}',
    )

    formats = formats_male + formats_female

    # link: http://dic.nicovideo.jp/a/日本人の名前一覧
    # link: http://www.meijiyasuda.co.jp/enjoy/ranking/
    first_names_female = (
        '明美', 'あすか', '香織', '加奈', 'くみ子', 'さゆり', '知実', '千代',
        '直子', '七夏', '花子', '春香', '真綾', '舞', '美加子', '幹',
        '桃子', '結衣', '裕美子', '陽子', '里佳',
    )

    # link: http://dic.nicovideo.jp/a/日本人の名前一覧
    # link: http://www.meijiyasuda.co.jp/enjoy/ranking/
    first_names_male = (
        '晃', '篤司', '治', '和也', '京助', '健一', '修平', '翔太',
        '淳', '聡太郎', '太一', '太郎', '拓真', '翼', '智也', '直樹',
        '直人', '英樹', '浩', '学', '充', '稔', '裕樹', '裕太',
        '康弘', '陽一', '洋介', '亮介', '涼平', '零',
    )

    first_names = first_names_male + first_names_female

    # link: https://dic.nicovideo.jp/a/日本の苗字(名字)の一覧
    last_names = (
        '青田', '青山', '石田', '井高', '伊藤', '井上', '宇野', '江古田',
        '大垣', '加藤', '加納', '喜嶋', '木村', '桐山', '工藤', '小泉',
        '小林', '近藤', '斉藤', '坂本', '佐々木', '佐藤', '笹田', '鈴木',
        '杉山', '高橋', '田中', '田辺', '津田', '中島', '中村', '渚',
        '中津川', '西之園', '野村', '原田', '浜田', '廣川', '藤本', '松本',
        '三宅', '宮沢', '村山', '山岸', '山口', '山田', '山本', '吉田',
        '吉本', '若松', '渡辺',
    )

    kana_formats = (
        '{{last_kana_name}} {{first_kana_name_female}}',
        '{{last_kana_name}} {{first_kana_name_male}}',
    )

    first_kana_names_female = (
        'アケミ', 'アスカ', 'カオリ', 'カナ',
        'クミコ', 'サユリ', 'サトミ', 'チヨ',
        'ナオコ', 'ナナミ', 'ハナコ', 'ハルカ',
        'マアヤ', 'マイ', 'ミカコ', 'ミキ',
        'モモコ', 'ユイ', 'ユミコ', 'ヨウコ',
        'リカ',
    )

    first_kana_names_male = (
        'アキラ', 'アツシ', 'オサム', 'カズヤ',
        'キョウスケ', 'ケンイチ', 'シュウヘイ', 'ショウタ',
        'ジュン', 'ソウタロウ', 'タイチ', 'タロウ',
        'タクマ', 'ツバサ', 'トモヤ', 'ナオキ',
        'ナオト', 'ヒデキ', 'ヒロシ', 'マナブ',
        'ミツル', 'ミノル', 'ユウキ', 'ユウタ',
        'ヤスヒロ', 'ヨウイチ', 'ヨウスケ', 'リョウスケ',
        'リョウヘイ', 'レイ',
    )

    first_kana_names = first_kana_names_male + first_kana_names_female

    last_kana_names = (
        'アオタ', 'アオヤマ', 'イシダ', 'イダカ',
        'イトウ', 'イノウエ', 'ウノ', 'エコダ',
        'オオガキ', 'カトウ', 'カノウ', 'キジマ',
        'キムラ', 'キリヤマ', 'クドウ', 'コイズミ',
        'コバヤシ', 'コンドウ', 'サイトウ', 'サカモト',
        'ササキ', 'サトウ', 'ササダ', 'スズキ',
        'スギヤマ', 'タカハシ', 'タナカ', 'タナベ',
        'ツダ', 'ナカジマ', 'ナカムラ', 'ナギサ',
        'ナカツガワ', 'ニシノソノ', 'ノムラ', 'ハラダ',
        'ハマダ', 'ヒロカワ', 'フジモト', 'マツモト',
        'ミヤケ', 'ミヤザワ', 'ムラヤマ', 'ヤマギシ',
        'ヤマグチ', 'ヤマダ', 'ヤマモト', 'ヨシダ',
        'ヨシモト', 'ワカマツ', 'ワタナベ',
    )

    romanized_formats = (
        '{{first_romanized_name_female}} {{last_romanized_name}}',
        '{{first_romanized_name_male}} {{last_romanized_name}}',
    )

    first_romanized_names_female = (
        'Akemi', 'Asuka', 'Kaori', 'Kana',
        'Kumiko', 'Sayuri', 'Satomi', 'Chiyo',
        'Naoko', 'Nanami', 'Hanako', 'Haruka',
        'Maaya', 'Mai', 'Mikako', 'Miki',
        'Momoko', 'Yui', 'Yumiko', 'Yoko',
        'Rika',
    )

    first_romanized_names_male = (
        'Akira', 'Atsushi', 'Osamu', 'Kazuya',
        'Kyosuke', 'Kenichi', 'Shohei', 'Shota',
        'Jun', 'Sotaro', 'Taichi', 'Taro',
        'Takuma', 'Tsubasa', 'Tomoya', 'Naoki',
        'Naoto', 'Hideki', 'Hiroshi', 'Manabu',
        'Mituru', 'Minoru', 'Yuki', 'Yuta',
        'Yasuhiro', 'Yoichi', 'Yosuke', 'Ryosuke',
        'Ryohei', 'Rei',
    )

    first_romanized_names = first_romanized_names_male \
        + first_romanized_names_female

    last_romanized_names = (
        'Aota', 'Aoyama', 'Ishida', 'Idaka',
        'Ito', 'Inoue', 'Uno', 'Ekoda',
        'Ogaki', 'Kato', 'Kano', 'Kijima',
        'Kimura', 'Kiriyama', 'Kudo', 'Koizumi',
        'Kobayashi', 'Kondo', 'Saito', 'Sakamoto',
        'Sasaki', 'Sato', 'Sasada', 'Suzuki',
        'Sugiyama', 'Takahashi', 'Tanaka', 'Tanabe',
        'Tsuda', 'Nakajima', 'Nakamura', 'Nagisa',
        'Nakatsugawa', 'Nishinosono', 'Nomura', 'Harada',
        'Hamada', 'Hirokawa', 'Fujimoto', 'Matsumoto',
        'Miyake', 'Miyagawa', 'Murayama', 'Yamagishi',
        'Yamaguchi', 'Yamada', 'Yamamoto', 'Yoshida',
        'Yoshimoto', 'Wakamatsu', 'Watanabe',
    )

    def kana_name(self):
        '''
        @example 'アオタ アキラ'
        '''
        pattern = self.random_element(self.kana_formats)
        return self.generator.parse(pattern)

    def first_kana_name(self):
        '''
        @example 'アキラ'
        '''
        return self.random_element(self.first_kana_names)

    def first_kana_name_female(self):
        return self.random_element(self.first_kana_names_female)

    def first_kana_name_male(self):
        return self.random_element(self.first_kana_names_male)

    def last_kana_name(self):
        '''
        @example 'アオタ'
        '''
        return self.random_element(self.last_kana_names)

    def romanized_name(self):
        '''
        @example 'Akira Aota'
        '''
        pattern = self.random_element(self.romanized_formats)
        return self.generator.parse(pattern)

    def first_romanized_name(self):
        '''
        @example 'Akira'
        '''
        return self.random_element(self.first_romanized_names)

    def first_romanized_name_female(self):
        return self.random_element(self.first_romanized_names_female)

    def first_romanized_name_male(self):
        return self.random_element(self.first_romanized_names_male)

    def last_romanized_name(self):
        '''
        @example 'Aota'
        '''
        return self.random_element(self.last_romanized_names)



import json
import time
from io import open

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage

from .exceptions import (
    WebpackError,
    WebpackLoaderBadStatsError,
    WebpackLoaderTimeoutError,
    WebpackBundleLookupError
)


class WebpackLoader(object):
    _assets = {}

    def __init__(self, name, config):
        self.name = name
        self.config = config

    def load_assets(self):
        try:
            with open(self.config['STATS_FILE'], encoding="utf-8") as f:
                return json.load(f)
        except IOError:
            raise IOError(
                'Error reading {0}. Are you sure webpack has generated '
                'the file and the path is correct?'.format(
                    self.config['STATS_FILE']))

    def get_assets(self):
        if self.config['CACHE']:
            if self.name not in self._assets:
                self._assets[self.name] = self.load_assets()
            return self._assets[self.name]
        return self.load_assets()

    def filter_chunks(self, chunks):
        for chunk in chunks:
            ignore = any(regex.match(chunk['name'])
                         for regex in self.config['ignores'])
            if not ignore:
                chunk['url'] = self.get_chunk_url(chunk)
                yield chunk

    def get_chunk_url(self, chunk):
        public_path = chunk.get('publicPath')
        if public_path:
            return public_path

        relpath = '{0}{1}'.format(
            self.config['BUNDLE_DIR_NAME'], chunk['name']
        )
        return staticfiles_storage.url(relpath)

    def get_bundle(self, bundle_name):
        assets = self.get_assets()

        # poll when debugging and block request until bundle is compiled
        # or the build times out
        if settings.DEBUG:
            timeout = self.config['TIMEOUT'] or 0
            timed_out = False
            start = time.time()
            while assets['status'] == 'compiling' and not timed_out:
                time.sleep(self.config['POLL_INTERVAL'])
                if timeout and (time.time() - timeout > start):
                    timed_out = True
                assets = self.get_assets()

            if timed_out:
                raise WebpackLoaderTimeoutError(
                    "Timed Out. Bundle `{0}` took more than {1} seconds "
                    "to compile.".format(bundle_name, timeout)
                )

        if assets.get('status') == 'done':
            chunks = assets['chunks'].get(bundle_name, None)
            if chunks is None:
                raise WebpackBundleLookupError('Cannot resolve bundle {0}.'.format(bundle_name))
            return self.filter_chunks(chunks)

        elif assets.get('status') == 'error':
            if 'file' not in assets:
                assets['file'] = ''
            if 'error' not in assets:
                assets['error'] = 'Unknown Error'
            if 'message' not in assets:
                assets['message'] = ''
            error = u"""
            {error} in {file}
            {message}
            """.format(**assets)
            raise WebpackError(error)

        raise WebpackLoaderBadStatsError(
            "The stats file does not contain valid data. Make sure "
            "webpack-bundle-tracker plugin is enabled and try to run "
            "webpack again.")



"""
Helpers to manipulate deferred DDL statements that might need to be adjusted or
discarded within when executing a migration.
"""


class Reference:
    """Base class that defines the reference interface."""

    def references_table(self, table):
        """
        Return whether or not this instance references the specified table.
        """
        return False

    def references_column(self, table, column):
        """
        Return whether or not this instance references the specified column.
        """
        return False

    def rename_table_references(self, old_table, new_table):
        """
        Rename all references to the old_name to the new_table.
        """
        pass

    def rename_column_references(self, table, old_column, new_column):
        """
        Rename all references to the old_column to the new_column.
        """
        pass

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))

    def __str__(self):
        raise NotImplementedError('Subclasses must define how they should be converted to string.')


class Table(Reference):
    """Hold a reference to a table."""

    def __init__(self, table, quote_name):
        self.table = table
        self.quote_name = quote_name

    def references_table(self, table):
        return self.table == table

    def rename_table_references(self, old_table, new_table):
        if self.table == old_table:
            self.table = new_table

    def __str__(self):
        return self.quote_name(self.table)


class TableColumns(Table):
    """Base class for references to multiple columns of a table."""

    def __init__(self, table, columns):
        self.table = table
        self.columns = columns

    def references_column(self, table, column):
        return self.table == table and column in self.columns

    def rename_column_references(self, table, old_column, new_column):
        if self.table == table:
            for index, column in enumerate(self.columns):
                if column == old_column:
                    self.columns[index] = new_column


class Columns(TableColumns):
    """Hold a reference to one or many columns."""

    def __init__(self, table, columns, quote_name, col_suffixes=()):
        self.quote_name = quote_name
        self.col_suffixes = col_suffixes
        super().__init__(table, columns)

    def __str__(self):
        def col_str(column, idx):
            try:
                return self.quote_name(column) + self.col_suffixes[idx]
            except IndexError:
                return self.quote_name(column)

        return ', '.join(col_str(column, idx) for idx, column in enumerate(self.columns))


class IndexName(TableColumns):
    """Hold a reference to an index name."""

    def __init__(self, table, columns, suffix, create_index_name):
        self.suffix = suffix
        self.create_index_name = create_index_name
        super().__init__(table, columns)

    def __str__(self):
        return self.create_index_name(self.table, self.columns, self.suffix)


class IndexColumns(Columns):
    def __init__(self, table, columns, quote_name, col_suffixes=(), opclasses=()):
        self.opclasses = opclasses
        super().__init__(table, columns, quote_name, col_suffixes)

    def __str__(self):
        def col_str(column, idx):
            # Index.__init__() guarantees that self.opclasses is the same
            # length as self.columns.
            col = '{} {}'.format(self.quote_name(column), self.opclasses[idx])
            try:
                col = '{} {}'.format(col, self.col_suffixes[idx])
            except IndexError:
                pass
            return col

        return ', '.join(col_str(column, idx) for idx, column in enumerate(self.columns))


class ForeignKeyName(TableColumns):
    """Hold a reference to a foreign key name."""

    def __init__(self, from_table, from_columns, to_table, to_columns, suffix_template, create_fk_name):
        self.to_reference = TableColumns(to_table, to_columns)
        self.suffix_template = suffix_template
        self.create_fk_name = create_fk_name
        super().__init__(from_table, from_columns,)

    def references_table(self, table):
        return super().references_table(table) or self.to_reference.references_table(table)

    def references_column(self, table, column):
        return (
            super().references_column(table, column) or
            self.to_reference.references_column(table, column)
        )

    def rename_table_references(self, old_table, new_table):
        super().rename_table_references(old_table, new_table)
        self.to_reference.rename_table_references(old_table, new_table)

    def rename_column_references(self, table, old_column, new_column):
        super().rename_column_references(table, old_column, new_column)
        self.to_reference.rename_column_references(table, old_column, new_column)

    def __str__(self):
        suffix = self.suffix_template % {
            'to_table': self.to_reference.table,
            'to_column': self.to_reference.columns[0],
        }
        return self.create_fk_name(self.table, self.columns, suffix)


class Statement(Reference):
    """
    Statement template and formatting parameters container.

    Allows keeping a reference to a statement without interpolating identifiers
    that might have to be adjusted if they're referencing a table or column
    that is removed
    """
    def __init__(self, template, **parts):
        self.template = template
        self.parts = parts

    def references_table(self, table):
        return any(
            hasattr(part, 'references_table') and part.references_table(table)
            for part in self.parts.values()
        )

    def references_column(self, table, column):
        return any(
            hasattr(part, 'references_column') and part.references_column(table, column)
            for part in self.parts.values()
        )

    def rename_table_references(self, old_table, new_table):
        for part in self.parts.values():
            if hasattr(part, 'rename_table_references'):
                part.rename_table_references(old_table, new_table)

    def rename_column_references(self, table, old_column, new_column):
        for part in self.parts.values():
            if hasattr(part, 'rename_column_references'):
                part.rename_column_references(table, old_column, new_column)

    def __str__(self):
        return self.template % self.parts



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("taggit", "0002_auto_20150616_2121"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="taggeditem", unique_together={("content_type", "object_id", "tag")}
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contenttypes", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        help_text="",
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="", unique=True, max_length=100, verbose_name="name"
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="", unique=True, max_length=100, verbose_name="slug"
                    ),
                ),
            ],
            options={"verbose_name": "tag", "verbose_name_plural": "tags"},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="TaggedItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        help_text="",
                        verbose_name="ID",
                    ),
                ),
                (
                    "object_id",
                    models.IntegerField(
                        help_text="", verbose_name="object ID", db_index=True
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        related_name="taggit_taggeditem_tagged_items",
                        verbose_name="content type",
                        to="contenttypes.ContentType",
                        help_text="",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "tag",
                    models.ForeignKey(
                        related_name="taggit_taggeditem_items",
                        to="taggit.Tag",
                        help_text="",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name": "tagged item",
                "verbose_name_plural": "tagged items",
            },
            bases=(models.Model,),
        ),
    ]



from datetime import tzinfo, timedelta


class TomlTz(tzinfo):
    def __init__(self, toml_offset):
        if toml_offset == "Z":
            self._raw_offset = "+00:00"
        else:
            self._raw_offset = toml_offset
        self._sign = -1 if self._raw_offset[0] == '-' else 1
        self._hours = int(self._raw_offset[1:3])
        self._minutes = int(self._raw_offset[4:6])

    def __deepcopy__(self, memo):
        return self.__class__(self._raw_offset)

    def tzname(self, dt):
        return "UTC" + self._raw_offset

    def utcoffset(self, dt):
        return self._sign * timedelta(hours=self._hours, minutes=self._minutes)

    def dst(self, dt):
        return timedelta(0)



"""Indirection for time functions.

We intentionally grab some "time" functions internally to avoid tests mocking "time" to affect
pytest runtime information (issue #185).

Fixture "mock_timinig" also interacts with this module for pytest's own tests.
"""
from time import perf_counter
from time import sleep
from time import time

__all__ = ["perf_counter", "sleep", "time"]



import os
import tempfile
from collections import namedtuple

from django.core.files import File
from django.core.files.storage import default_storage

try:
    from PIL import Image, ImageFile
except ImportError:
    import Image
    import ImageFile

ImageFile.MAXBLOCK = 1024 * 1024

from . import settings, signals, utils


class CropSpec(
    namedtuple(
        "CropSpec", ["name", "editable", "width", "height", "x", "x2", "y", "y2"]
    )
):
    def to_dict(self):
        return self._asdict()


class CropConfig:
    def __init__(
        self, name, width=None, height=None, quality=90, editable=True, upscale=False
    ):
        self.name = name
        self.width = width
        self.height = height
        self.upscale = upscale
        self.quality = quality
        self.editable = editable

    def _adjust_coordinates(self, ratio, current_size, needed_size):
        if not ratio:
            ratio = 1
        diff = current_size - (needed_size/ratio)
        if diff:
            adjust = int(diff/2)
        else:
            adjust = 0
        return adjust

    def get_crop_spec(self, im, x=None, x2=None, y=None, y2=None):
        """
        Returns the default crop points for this image.
        """
        w, h = [float(v) for v in im.size]
        upscale = self.upscale
        if x is not None and x2 and y is not None and y2:
            upscale = True
            w = float(x2) - x
            h = float(y2) - y
        else:
            x = 0
            x2 = w
            y = 0
            y2 = h

        if self.width and self.height:
            ry = self.height/h
            rx = self.width/w
            if rx < ry:
                ratio = ry
                adjust = self._adjust_coordinates(ratio, w, self.width)
                x = x + adjust
                x2 = x2 - adjust
            else:
                ratio = rx
                adjust = self._adjust_coordinates(ratio, h, self.height)
                y = y + adjust
                y2 = y2 - adjust

            width = self.width
            height = self.height
        elif self.width:
            ratio = self.width/w
            width = self.width
            height = int(h * ratio)
        else:
            ratio = self.height/h
            width = int(w * ratio)
            height = self.height

        if ratio > 1 and not upscale:
            return

        # A point can't have a negative value 
        x, x2, y, y2 = max(0,int(x)), max(0,int(x2)), max(0,int(y)), max(0,int(y2))
        return CropSpec(
            name=self.name,
            editable=self.editable,
            width=width,
            height=height,
            x=x,
            x2=x2,
            y=y,
            y2=y2,
        )

    def rotate_by_exif(self, im):
        if "exif" in im.info:
            try:
                exifinfo = im._getexif()
            except (IOError, KeyError, IndexError):
                return im

            if exifinfo is not None:
                # 274 is exif code for orientation
                orientation = exifinfo.get(274, None)
                if orientation == 3:
                    # flip 180
                    im = im.rotate(180)
                elif orientation == 6:
                    # flip -90
                    im = im.rotate(-90)
                elif orientation == 8:
                    # flip 90
                    im = im.rotate(90)
        return im

    def process_image(self, im, crop_spec=None):
        if not crop_spec:
            crop_spec = self.get_crop_spec(im)

        if crop_spec:
            return scale_and_crop(im, crop_spec)


class Cropper:
    _registry = {}

    def __init__(self, storage=None):
        self.storage = storage or default_storage
        self._required_crops = []
        for k, v in list(settings.IMAGE_SIZES.items()):
            if v and isinstance(v, dict):
                self.register(CropConfig(k, **v))
                self._required_crops.append(k)

    def get_crop_config(self, name):
        return self._registry.get(name)

    def required_crops(self):
        return self._required_crops

    def register(self, config):
        assert isinstance(config, CropConfig)
        if config.name in self._registry:
            seen = self._registry[config.name]
            for attr in ("name", "width", "height", "upscale", "quality", "editable"):
                if getattr(seen, attr) != getattr(config, attr):
                    raise ValueError(
                        "{0} is already registered as a different crop".format(
                            config.name
                        )
                    )
        self._registry[config.name] = config

    def unregister(self, name):
        if name in self._registry:
            del self._registry[name]

    def create_crop(self, name, file_obj, x=None, x2=None, y=None, y2=None):
        """
        Generate Version for an Image.
        value has to be a serverpath relative to MEDIA_ROOT.

        Returns the spec for the crop that was created.
        """

        if name not in self._registry:
            return

        file_obj.seek(0)
        im = Image.open(file_obj)
        config = self._registry[name]

        if x is not None and x2 and y is not None and y2 and not config.editable:
            # You can't ask for something special
            # for non editable images
            return

        im = config.rotate_by_exif(im)
        crop_spec = config.get_crop_spec(im, x=x, x2=x2, y=y, y2=y2)
        image = config.process_image(im, crop_spec=crop_spec)
        if image:
            crop_name = utils.get_size_filename(file_obj.name, name)
            self._save_file(image, crop_name)
            return crop_spec

    def replace_image(self, file_obj, x=None, x2=None, y=None, y2=None):
        assert x is not None and x2 and y is not None and y2

        file_obj.seek(0)
        im = Image.open(file_obj)

        crop_spec = CropSpec(
            x=x,
            x2=x2,
            y=y,
            y2=y2,
            width=x2 - x,
            height=y2 - y,
            name="original",
            editable=False,
        )
        image = scale_and_crop(im, crop_spec=crop_spec)
        if image:
            self._save_file(image, file_obj.name)
            return crop_spec

    def _save_file(self, im, filename, quality=90):
        tmpfile = File(tempfile.NamedTemporaryFile())
        try:
            root, ext = os.path.splitext(filename)
            try:
                im.save(
                    tmpfile,
                    format=Image.EXTENSION[ext.lower()],
                    quality=quality,
                    optimize=(ext != ".gif"),
                )
            except IOError:
                im.save(tmpfile, format=Image.EXTENSION[ext.lower()], quality=quality)

            if filename != self.storage.get_available_name(filename):
                self.storage.delete(filename)

            self.storage.save(filename, tmpfile)
            url = self.storage.url(filename)
            utils.update_cache_bust_version(url)
            signals.file_saved.send(filename)
        finally:
            tmpfile.close()


def scale_and_crop(im, crop_spec):
    """
    Scale and Crop.
    """
    im = im.crop((crop_spec.x, crop_spec.y, crop_spec.x2, crop_spec.y2))

    if crop_spec.width and crop_spec.height:
        im = im.resize((crop_spec.width, crop_spec.height), resample=Image.ANTIALIAS)

    return im


cropper = Cropper()



from .. import Provider as SsnProvider


def checksum(digits):
    """
    Returns the checksum of CPF digits.
    References to the algorithm:
    https://pt.wikipedia.org/wiki/Cadastro_de_pessoas_f%C3%ADsicas#Algoritmo
    https://metacpan.org/source/MAMAWE/Algorithm-CheckDigits-v1.3.0/lib/Algorithm/CheckDigits/M11_004.pm
    """
    s = 0
    p = len(digits) + 1
    for i in range(0, len(digits)):
        s += digits[i] * p
        p -= 1

    reminder = s % 11
    if reminder == 0 or reminder == 1:
        return 0
    else:
        return 11 - reminder


class Provider(SsnProvider):
    """
    Provider for Brazilian SSN also known in Brazil as CPF.
    There are two methods Provider.ssn and Provider.cpf
    The snn returns a valid number with numbers only
    The cpf return a valid number formatted with brazilian mask. eg nnn.nnn.nnn-nn
    """

    def ssn(self):
        digits = self.generator.random.sample(range(10), 9)

        dv = checksum(digits)
        digits.append(dv)
        digits.append(checksum(digits))

        return ''.join(map(str, digits))

    def cpf(self):
        c = self.ssn()
        return c[:3] + '.' + c[3:6] + '.' + c[6:9] + '-' + c[9:]

    def rg(self):
        """
        Brazilian RG, return plain numbers.
        Check:  https://www.ngmatematica.com/2014/02/como-determinar-o-digito-verificador-do.html
        """

        digits = self.generator.random.sample(range(0, 9), 8)
        checksum = sum(i * digits[i - 2] for i in range(2, 10))
        last_digit = 11 - (checksum % 11)

        if last_digit == 10:
            digits.append('X')
        elif last_digit == 11:
            digits.append(0)
        else:
            digits.append(last_digit)

        return ''.join(map(str, digits))



from .manager import cache_manager

default_app_config = "scarlet.cache.apps.AppConfig"


def autodiscover():
    """
    Copied from django source

    Auto-discover INSTALLED_APPS cache_groups.py modules and fail silently if
    not present. This forces an import on them to register any admin bits they
    may want.
    """

    import copy

    from django.conf import settings
    from django.utils.importlib import import_module
    from django.utils.module_loading import module_has_submodule

    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's admin module.
        try:
            before_import_registry = copy.copy(cache_manager._registry)
            import_module(f"{app}.cache_manager")
        except Exception:
            # Reset the model registry to the state before the last import as
            # this import will have to reoccur on the next request and this
            # could raise NotRegistered and AlreadyRegistered exceptions
            # (see #8245).
            cache_manager._registry = before_import_registry

            # Decide whether to bubble up this error. If the app just
            # doesn't have an admin module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, "cache_manager"):
                raise

default_app_config = "scarlet.cache.apps.AppConfig"



from django.conf import settings as django_settings

INTERNAL_TAG_HANDLER = getattr(
    django_settings, "SCARLET_INTERNAL_TAG_HANDLER", ".taggit_handler"
)

USE_SCARLET_DATE_FORMATS = getattr(django_settings, "USE_SCARLET_DATE_FORMATS", True)
DATETIME_INPUT_FORMATS = getattr(
    django_settings, "SCARLET_DATETIME_INPUT_FORMAT", ["%m/%d/%Y %I:%M:%S %p",],
)
DATE_INPUT_FORMATS = getattr(django_settings, "DATE_INPUT_FORMT", "%m/%d/%Y",)



import pprint
import reprlib
from typing import Any
from typing import Dict
from typing import IO
from typing import Optional


def _try_repr_or_str(obj: object) -> str:
    try:
        return repr(obj)
    except (KeyboardInterrupt, SystemExit):
        raise
    except BaseException:
        return '{}("{}")'.format(type(obj).__name__, obj)


def _format_repr_exception(exc: BaseException, obj: object) -> str:
    try:
        exc_info = _try_repr_or_str(exc)
    except (KeyboardInterrupt, SystemExit):
        raise
    except BaseException as exc:
        exc_info = "unpresentable exception ({})".format(_try_repr_or_str(exc))
    return "<[{} raised in repr()] {} object at 0x{:x}>".format(
        exc_info, type(obj).__name__, id(obj)
    )


def _ellipsize(s: str, maxsize: int) -> str:
    if len(s) > maxsize:
        i = max(0, (maxsize - 3) // 2)
        j = max(0, maxsize - 3 - i)
        return s[:i] + "..." + s[len(s) - j :]
    return s


class SafeRepr(reprlib.Repr):
    """repr.Repr that limits the resulting size of repr() and includes
    information on exceptions raised during the call."""

    def __init__(self, maxsize: int) -> None:
        super().__init__()
        self.maxstring = maxsize
        self.maxsize = maxsize

    def repr(self, x: object) -> str:
        try:
            s = super().repr(x)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exc:
            s = _format_repr_exception(exc, x)
        return _ellipsize(s, self.maxsize)

    def repr_instance(self, x: object, level: int) -> str:
        try:
            s = repr(x)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exc:
            s = _format_repr_exception(exc, x)
        return _ellipsize(s, self.maxsize)


def safeformat(obj: object) -> str:
    """Return a pretty printed string for the given object.

    Failing __repr__ functions of user instances will be represented
    with a short exception info.
    """
    try:
        return pprint.pformat(obj)
    except Exception as exc:
        return _format_repr_exception(exc, obj)


def saferepr(obj: object, maxsize: int = 240) -> str:
    """Return a size-limited safe repr-string for the given object.

    Failing __repr__ functions of user instances will be represented
    with a short exception info and 'saferepr' generally takes
    care to never raise exceptions itself.

    This function is a wrapper around the Repr/reprlib functionality of the
    standard 2.6 lib.
    """
    return SafeRepr(maxsize).repr(obj)


class AlwaysDispatchingPrettyPrinter(pprint.PrettyPrinter):
    """PrettyPrinter that always dispatches (regardless of width)."""

    def _format(
        self,
        object: object,
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Dict[int, Any],
        level: int,
    ) -> None:
        # Type ignored because _dispatch is private.
        p = self._dispatch.get(type(object).__repr__, None)  # type: ignore[attr-defined]

        objid = id(object)
        if objid in context or p is None:
            # Type ignored because _format is private.
            super()._format(  # type: ignore[misc]
                object, stream, indent, allowance, context, level,
            )
            return

        context[objid] = 1
        p(self, object, stream, indent, allowance, context, level + 1)
        del context[objid]


def _pformat_dispatch(
    object: object,
    indent: int = 1,
    width: int = 80,
    depth: Optional[int] = None,
    *,
    compact: bool = False
) -> str:
    return AlwaysDispatchingPrettyPrinter(
        indent=indent, width=width, depth=depth, compact=compact
    ).pformat(object)



# coding: utf-8
from __future__ import unicode_literals

import re
import json
import pickle
import textwrap
import unittest
import importlib
import importlib_metadata
import pyfakefs.fake_filesystem_unittest as ffs

from . import fixtures
from .. import (
    Distribution, EntryPoint, MetadataPathFinder,
    PackageNotFoundError, distributions,
    entry_points, metadata, version,
    )

try:
    from builtins import str as text
except ImportError:
    from __builtin__ import unicode as text


class BasicTests(fixtures.DistInfoPkg, unittest.TestCase):
    version_pattern = r'\d+\.\d+(\.\d)?'

    def test_retrieves_version_of_self(self):
        dist = Distribution.from_name('distinfo-pkg')
        assert isinstance(dist.version, text)
        assert re.match(self.version_pattern, dist.version)

    def test_for_name_does_not_exist(self):
        with self.assertRaises(PackageNotFoundError):
            Distribution.from_name('does-not-exist')

    def test_package_not_found_mentions_metadata(self):
        """
        When a package is not found, that could indicate that the
        packgae is not installed or that it is installed without
        metadata. Ensure the exception mentions metadata to help
        guide users toward the cause. See #124.
        """
        with self.assertRaises(PackageNotFoundError) as ctx:
            Distribution.from_name('does-not-exist')

        assert "metadata" in str(ctx.exception)

    def test_new_style_classes(self):
        self.assertIsInstance(Distribution, type)
        self.assertIsInstance(MetadataPathFinder, type)


class ImportTests(fixtures.DistInfoPkg, unittest.TestCase):
    def test_import_nonexistent_module(self):
        # Ensure that the MetadataPathFinder does not crash an import of a
        # non-existent module.
        with self.assertRaises(ImportError):
            importlib.import_module('does_not_exist')

    def test_resolve(self):
        entries = dict(entry_points()['entries'])
        ep = entries['main']
        self.assertEqual(ep.load().__name__, "main")

    def test_entrypoint_with_colon_in_name(self):
        entries = dict(entry_points()['entries'])
        ep = entries['ns:sub']
        self.assertEqual(ep.value, 'mod:main')

    def test_resolve_without_attr(self):
        ep = EntryPoint(
            name='ep',
            value='importlib_metadata',
            group='grp',
            )
        assert ep.load() is importlib_metadata


class NameNormalizationTests(
        fixtures.OnSysPath, fixtures.SiteDir, unittest.TestCase):
    @staticmethod
    def pkg_with_dashes(site_dir):
        """
        Create minimal metadata for a package with dashes
        in the name (and thus underscores in the filename).
        """
        metadata_dir = site_dir / 'my_pkg.dist-info'
        metadata_dir.mkdir()
        metadata = metadata_dir / 'METADATA'
        with metadata.open('w') as strm:
            strm.write('Version: 1.0\n')
        return 'my-pkg'

    def test_dashes_in_dist_name_found_as_underscores(self):
        """
        For a package with a dash in the name, the dist-info metadata
        uses underscores in the name. Ensure the metadata loads.
        """
        pkg_name = self.pkg_with_dashes(self.site_dir)
        assert version(pkg_name) == '1.0'

    @staticmethod
    def pkg_with_mixed_case(site_dir):
        """
        Create minimal metadata for a package with mixed case
        in the name.
        """
        metadata_dir = site_dir / 'CherryPy.dist-info'
        metadata_dir.mkdir()
        metadata = metadata_dir / 'METADATA'
        with metadata.open('w') as strm:
            strm.write('Version: 1.0\n')
        return 'CherryPy'

    def test_dist_name_found_as_any_case(self):
        """
        Ensure the metadata loads when queried with any case.
        """
        pkg_name = self.pkg_with_mixed_case(self.site_dir)
        assert version(pkg_name) == '1.0'
        assert version(pkg_name.lower()) == '1.0'
        assert version(pkg_name.upper()) == '1.0'


class NonASCIITests(fixtures.OnSysPath, fixtures.SiteDir, unittest.TestCase):
    @staticmethod
    def pkg_with_non_ascii_description(site_dir):
        """
        Create minimal metadata for a package with non-ASCII in
        the description.
        """
        metadata_dir = site_dir / 'portend.dist-info'
        metadata_dir.mkdir()
        metadata = metadata_dir / 'METADATA'
        with metadata.open('w', encoding='utf-8') as fp:
            fp.write('Description: pôrˈtend\n')
        return 'portend'

    @staticmethod
    def pkg_with_non_ascii_description_egg_info(site_dir):
        """
        Create minimal metadata for an egg-info package with
        non-ASCII in the description.
        """
        metadata_dir = site_dir / 'portend.dist-info'
        metadata_dir.mkdir()
        metadata = metadata_dir / 'METADATA'
        with metadata.open('w', encoding='utf-8') as fp:
            fp.write(textwrap.dedent("""
                Name: portend

                pôrˈtend
                """).lstrip())
        return 'portend'

    def test_metadata_loads(self):
        pkg_name = self.pkg_with_non_ascii_description(self.site_dir)
        meta = metadata(pkg_name)
        assert meta['Description'] == 'pôrˈtend'

    def test_metadata_loads_egg_info(self):
        pkg_name = self.pkg_with_non_ascii_description_egg_info(self.site_dir)
        meta = metadata(pkg_name)
        assert meta.get_payload() == 'pôrˈtend\n'


class DiscoveryTests(fixtures.EggInfoPkg,
                     fixtures.DistInfoPkg,
                     unittest.TestCase):

    def test_package_discovery(self):
        dists = list(distributions())
        assert all(
            isinstance(dist, Distribution)
            for dist in dists
            )
        assert any(
            dist.metadata['Name'] == 'egginfo-pkg'
            for dist in dists
            )
        assert any(
            dist.metadata['Name'] == 'distinfo-pkg'
            for dist in dists
            )

    def test_invalid_usage(self):
        with self.assertRaises(ValueError):
            list(distributions(context='something', name='else'))


class DirectoryTest(fixtures.OnSysPath, fixtures.SiteDir, unittest.TestCase):
    def test_egg_info(self):
        # make an `EGG-INFO` directory that's unrelated
        self.site_dir.joinpath('EGG-INFO').mkdir()
        # used to crash with `IsADirectoryError`
        with self.assertRaises(PackageNotFoundError):
            version('unknown-package')

    def test_egg(self):
        egg = self.site_dir.joinpath('foo-3.6.egg')
        egg.mkdir()
        with self.add_sys_path(egg):
            with self.assertRaises(PackageNotFoundError):
                version('foo')


class MissingSysPath(fixtures.OnSysPath, unittest.TestCase):
    site_dir = '/does-not-exist'

    def test_discovery(self):
        """
        Discovering distributions should succeed even if
        there is an invalid path on sys.path.
        """
        importlib_metadata.distributions()


class InaccessibleSysPath(fixtures.OnSysPath, ffs.TestCase):
    site_dir = '/access-denied'

    def setUp(self):
        super(InaccessibleSysPath, self).setUp()
        self.setUpPyfakefs()
        self.fs.create_dir(self.site_dir, perm_bits=000)

    def test_discovery(self):
        """
        Discovering distributions should succeed even if
        there is an invalid path on sys.path.
        """
        list(importlib_metadata.distributions())


class TestEntryPoints(unittest.TestCase):
    def __init__(self, *args):
        super(TestEntryPoints, self).__init__(*args)
        self.ep = importlib_metadata.EntryPoint('name', 'value', 'group')

    def test_entry_point_pickleable(self):
        revived = pickle.loads(pickle.dumps(self.ep))
        assert revived == self.ep

    def test_immutable(self):
        """EntryPoints should be immutable"""
        with self.assertRaises(AttributeError):
            self.ep.name = 'badactor'

    def test_repr(self):
        assert 'EntryPoint' in repr(self.ep)
        assert 'name=' in repr(self.ep)
        assert "'name'" in repr(self.ep)

    def test_hashable(self):
        """EntryPoints should be hashable"""
        hash(self.ep)

    def test_json_dump(self):
        """
        json should not expect to be able to dump an EntryPoint
        """
        with self.assertRaises(Exception):
            json.dumps(self.ep)

    def test_module(self):
        assert self.ep.module == 'value'

    def test_attr(self):
        assert self.ep.attr is None


class FileSystem(
        fixtures.OnSysPath, fixtures.SiteDir, fixtures.FileBuilder,
        unittest.TestCase):
    def test_unicode_dir_on_sys_path(self):
        """
        Ensure a Unicode subdirectory of a directory on sys.path
        does not crash.
        """
        fixtures.build_files(
            {self.unicode_filename(): {}},
            prefix=self.site_dir,
            )
        list(distributions())



from importlib import import_module
from django.conf import settings
from .config import load_config


_loaders = {}


def import_string(dotted_path):
    '''
    This is a rough copy of django's import_string, which wasn't introduced until Django 1.7

    Once this package's support for Django 1.6 has been removed, this can be safely replaced with
    `from django.utils.module_loading import import_string`
    '''
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ValueError, AttributeError, ImportError):
        raise ImportError('%s doesn\'t look like a valid module path' % dotted_path)


def get_loader(config_name):
    if config_name not in _loaders:
        config = load_config(config_name)
        loader_class = import_string(config['LOADER_CLASS'])
        _loaders[config_name] = loader_class(config_name, config)
    return _loaders[config_name]


def _filter_by_extension(bundle, extension):
    '''Return only files with the given extension'''
    for chunk in bundle:
        if chunk['name'].endswith('.{0}'.format(extension)):
            yield chunk


def _get_bundle(bundle_name, extension, config):
    bundle = get_loader(config).get_bundle(bundle_name)
    if extension:
        bundle = _filter_by_extension(bundle, extension)
    return bundle


def get_files(bundle_name, extension=None, config='DEFAULT'):
    '''Returns list of chunks from named bundle'''
    return list(_get_bundle(bundle_name, extension, config))


def get_as_tags(bundle_name, extension=None, config='DEFAULT', attrs=''):
    '''
    Get a list of formatted <script> & <link> tags for the assets in the
    named bundle.

    :param bundle_name: The name of the bundle
    :param extension: (optional) filter by extension, eg. 'js' or 'css'
    :param config: (optional) the name of the configuration
    :return: a list of formatted tags as strings
    '''

    bundle = _get_bundle(bundle_name, extension, config)
    tags = []
    for chunk in bundle:
        if chunk['name'].endswith(('.js', '.js.gz')):
            tags.append((
                '<script type="text/javascript" src="{0}" {1}></script>'
            ).format(chunk['url'], attrs))
        elif chunk['name'].endswith(('.css', '.css.gz')):
            tags.append((
                '<link type="text/css" href="{0}" rel="stylesheet" {1}/>'
            ).format(chunk['url'], attrs))
    return tags


def get_static(asset_name, config='DEFAULT'):
    '''
    Equivalent to Django's 'static' look up but for webpack assets.

    :param asset_name: the name of the asset
    :param config: (optional) the name of the configuration
    :return: path to webpack asset as a string
    '''
    return "{0}{1}".format(
        get_loader(config).get_assets().get(
            'publicPath', getattr(settings, 'STATIC_URL')
        ),
        asset_name
    )
